"""Utilities for working with 3D vectors in cartesian space.

note: some of these functions are redundant as they are provided by built-in numpy operations.

a vector is a one dimensional numpy array with 3 elements: x, y, z.
some functions accept a tuple or list of 3 elements as an alternative to a numpy array.
"""

import logging

log = logging.getLogger(__name__)

import math as maths
import numpy as np
import numba  # type: ignore
from numba import njit  # type: ignore
from typing import Tuple, Optional


def radians_from_degrees(deg):
    """Converts angle from degrees to radians."""
    return np.radians(deg)


def degrees_from_radians(rad):
    """Converts angle from radians to degrees."""
    return np.degrees(rad)


def zero_vector():
    """Returns a zero vector [0.0, 0.0, 0.0]."""
    return np.zeros(3)


def v_3d(v):
    """Returns a 3D vector for a 2D or 3D vector."""
    assert 2 <= len(v) <= 3
    if len(v) == 3:
        return v
    v3 = np.zeros(3)
    v3[:2] = v
    return v3


def add(a, b):  # note: could just use numpy a + b facility
    """Returns vector sum a+b."""
    a = np.array(a)
    b = np.array(b)
    assert a.size == b.size
    return a + b


def subtract(a, b):  # note: could just use numpy a - b facility
    """Returns vector difference a-b."""
    a = np.array(a)
    b = np.array(b)
    assert a.size == b.size
    return a - b


def elemental_multiply(a, b):  # note: could just use numpy a * b facility
    """Returns vector with products of corresponding elements of a and b."""
    a = np.array(a)
    b = np.array(b)
    assert a.size == b.size
    return a * b


def amplify(v, scaling):  # note: could just use numpy a * scalar facility
    """Returns vector with direction of v, amplified by scaling."""
    v = np.array(v)
    return scaling * v


def unit_vector(v):
    """Returns vector with same direction as v but with unit length."""
    assert 2 <= len(v) <= 3
    v = np.array(v, dtype = np.float64)
    norm = np.linalg.norm(v)
    if norm == 0.0:
        return v
    v = v / norm
    return v


@njit
def unit_vector_njit(v):  # pragma: no cover
    """Returns vector with same direction as v but with unit length."""
    norm = np.linalg.norm(v)
    if norm == 0.0:
        return v
    v = v / norm
    return v


def unit_vectors(v):
    """Returns vectors with same direction as those in v but with unit length."""
    scaling = np.sqrt(np.sum(v * v, axis = -1))
    zero_mask = np.zeros(v.shape, dtype = bool)
    zero_mask[np.where(scaling == 0.0), :] = True
    restore = np.seterr(all = 'ignore')
    result = np.where(zero_mask, 0.0, v / np.expand_dims(scaling, -1))
    np.seterr(**restore)
    return result


def nan_unit_vectors(v):
    """Returns vectors with same direction as those in v but with unit length, allowing NaNs."""
    assert v.shape[-1] == 3
    vf = v.reshape((-1, 3))
    nan_mask = np.isnan(vf)
    restore = np.seterr(all = 'ignore')
    scaling = np.sqrt(np.sum(vf * vf, axis = -1))
    zero_mask = np.zeros(vf.shape, dtype = bool)
    zero_mask[np.where(scaling == 0.0), :] = True
    result = np.where(zero_mask, 0.0, vf / np.expand_dims(scaling, -1))
    result = np.where(nan_mask, np.nan, result)
    np.seterr(**restore)
    return result.reshape(v.shape)


def unit_vector_from_azimuth(azimuth):
    """Returns horizontal unit vector in compass bearing given by azimuth (x = East, y = North)."""
    azimuth = azimuth % 360.0
    azimuth_radians = radians_from_degrees(azimuth)
    result = zero_vector()
    result[0] = maths.sin(azimuth_radians)  # x (increasing to east)
    result[1] = maths.cos(azimuth_radians)  # y (increasing to north)
    return result  # leave z as zero


def azimuth(v):  # 'azimuth' is synonymous with 'compass bearing'
    """Returns the compass bearing in degrees of the direction of v (x = East, y = North), ignoring z."""
    assert 2 <= v.size <= 3
    z_zero_v = np.zeros(3)
    z_zero_v[:2] = v[:2]
    unit_v = unit_vector(z_zero_v)  # also checks that z_zero_v is not zero vector
    x = unit_v[0]
    y = unit_v[1]  # ignore z component
    if x == 0.0 and y == 0.0:
        return 0.0  # arbitrary azimuth of a vertical vector
    if abs(x) >= abs(y):
        radians = maths.pi / 2.0 - maths.atan(y / x)
        if x < 0.0:
            radians += maths.pi
    else:
        radians = maths.atan(x / y)
        if y < 0.0:
            radians += maths.pi
    if radians < 0.0:
        radians += 2.0 * maths.pi
    return degrees_from_radians(radians)


def azimuths(va):  # 'azimuth' is synonymous with 'compass bearing'
    """Returns the compass bearings in degrees of the direction of each vector in va (x = East, y = North), ignoring z."""
    assert va.ndim > 1 and 2 <= va.shape[-1] <= 3
    shape = tuple(list(va.shape[:-1]) + [3])
    z_zero_v = np.zeros(shape)
    z_zero_v[..., :2] = va[..., :2]
    unit_v = unit_vectors(z_zero_v)  # also checks that z_zero_v is not zero vector
    x = unit_v[..., 0]
    y = unit_v[..., 1]  # ignore z component
    # todo: handle cases where x == y == 0
    restore = np.seterr(all = 'ignore')
    radians = np.where(
        np.abs(x) >= np.abs(y),
        np.where(x < 0.0, maths.pi * 3.0 / 2.0 - np.arctan(y / x), maths.pi / 2.0 - np.arctan(y / x)),
        np.where(y < 0.0, maths.pi + np.arctan(x / y), np.arctan(x / y)))
    radians = radians % (2.0 * maths.pi)
    np.seterr(**restore)
    return np.degrees(radians)


def inclination(v):
    """Returns the inclination in degrees of v (angle relative to +ve z axis)."""
    assert len(v) == 3
    unit_v = unit_vector(v)
    radians = maths.acos(unit_v[2])
    return degrees_from_radians(radians)


def inclinations(a):
    """Returns the inclination in degrees of each vector in a (angle relative to +ve z axis)."""
    assert a.ndim > 1 and a.shape[-1] == 3
    unit_vs = unit_vectors(a)
    radians = np.arccos(unit_vs[..., 2])
    return degrees_from_radians(radians)


def nan_inclinations(a, already_unit_vectors = False):
    """Returns the inclination in degrees of each vector in a (angle relative to +ve z axis), allowing NaNs."""
    assert a.ndim > 1 and a.shape[-1] == 3
    unit_vs = a if already_unit_vectors else nan_unit_vectors(a)
    restore = np.seterr(all = 'ignore')
    radians = np.where(np.isnan(unit_vs[..., 2]), np.nan, np.arccos(unit_vs[..., 2]))
    result = np.where(np.isnan(radians), np.nan, degrees_from_radians(radians))
    np.seterr(**restore)
    return result


def points_direction_vector(a, axis):
    """Returns an average direction vector based on first and last non-NaN points or slices in given axis."""
    # note: as currently coded, might give poor results with some patterns of NaNs

    assert a.ndim > 1 and 0 <= axis < a.ndim - 1 and a.shape[-1] > 1 and a.shape[axis] > 1
    if np.all(np.isnan(a)):
        return None
    start = 0
    start_slicing = [slice(None)] * a.ndim
    while True:
        start_slicing[axis] = slice(start, start + 1)
        if not np.all(np.isnan(a[tuple(start_slicing)])):
            break
        start += 1
        assert start < a.shape[axis]
    finish = a.shape[axis] - 1
    finish_slicing = [slice(None)] * a.ndim
    while True:
        finish_slicing[axis] = slice(finish, finish + 1)
        if not np.all(np.isnan(a[tuple(finish_slicing)])):
            break
        finish -= 1
    if start >= finish:
        return None
    if a.ndim > 2:
        mean_axes = tuple(range(a.ndim - 1))
        start_p = np.nanmean(a[tuple(start_slicing)], axis = mean_axes)
        finish_p = np.nanmean(a[tuple(finish_slicing)], axis = mean_axes)
    else:
        start_p = a[start]
        finish_p = a[finish]

    return finish_p - start_p


def dot_product(a, b):  # pragma: no cover
    """Returns the dot product (scalar product) of the two vectors."""
    return np.dot(a, b)


def dot_products(a, b):  # pragma: no cover
    """Returns the dot products of pairs of vectors; last axis covers element of a vector."""
    return np.sum(a * b, axis = -1)


def cross_product(a, b):  # pragma: no cover
    """Returns the cross product (vector product) of the two vectors."""
    return np.cross(a, b)


def naive_length(v):  # pragma: no cover
    """Returns the length of the vector assuming consistent units."""
    return np.linalg.norm(v)


def naive_lengths(v):
    """Returns the lengths of the vectors assuming consistent units."""
    return np.sqrt(np.sum(v * v, axis = -1))


def naive_2d_length(v):  # pragma: no cover
    """Returns the length of the vector projected onto xy plane, assuming consistent units."""
    return np.linalg.norm(v[0:2])


def naive_2d_lengths(v):
    """Returns the lengths of the vectors projected onto xy plane, assuming consistent units."""
    v2d = v[..., :2]
    return np.sqrt(np.sum(v2d * v2d, axis = -1))


def unit_corrected_length(v, unit_conversion):
    """Returns the length of the vector v after applying the unit_conversion factors.

    arguments:
        v (1D numpy float array): vector with mixed units of measure
        unit_conversion (1D numpy float array): vector to multiply elements of v by, prior to finding length

    returns:
        float, being the length of v after adjustment by unit_conversion

    notes:
        example unit_conversion might be:
        [1.0, 1.0, 0.3048] to convert z from feet to metres, or
        [3.28084, 3.28084, 1.0] to convert x and y from metres to feet
    """
    converted = elemental_multiply(v, unit_conversion)
    return naive_length(converted)


def manhatten_distance(p1, p2):
    """Returns the Manhattan distance between two points."""
    return abs(p2[0] - p1[0]) + abs(p2[1] - p1[1]) + abs(p2[2] - p1[2])


def manhattan_distance(p1, p2):  #  alternative spelling to above
    """Returns the Manhattan distance between two points."""
    return abs(p2[0] - p1[0]) + abs(p2[1] - p1[1]) + abs(p2[2] - p1[2])


def radians_difference(a, b):
    """Returns the angle between two vectors, in radians."""

    return maths.acos(min(1.0, max(-1.0, dot_product(unit_vector(a), unit_vector(b)))))


def degrees_difference(a, b):
    """Returns the angle between two vectors, in degrees."""

    return degrees_from_radians(radians_difference(a, b))


def rotation_matrix_3d_axial(axis, angle):
    """Returns a rotation matrix which will rotate points about axis (0: x, 1: y, or 2: z) by angle in degrees.

    note:
       this function follows the mathematical convention: a positive angle results in anti-clockwise rotation
       when viewed in direction of positive axis
    """
    axis_a = (axis + 1) % 3
    axis_b = (axis_a + 1) % 3
    matrix = np.eye(3)
    radians = np.radians(angle)
    cosine = np.cos(radians)
    sine = np.sin(radians)
    matrix[axis_a, axis_a] = cosine
    matrix[axis_b, axis_b] = cosine
    matrix[axis_a, axis_b] = -sine  # left handed coordinate system, eg. UTM & depth
    matrix[axis_b, axis_a] = sine
    return matrix


def no_rotation_matrix():  # pragma: no cover
    """Returns a rotation matrix which will not move points (identity matrix)."""
    return np.eye(3)


def rotation_3d_matrix(xzy_axis_angles):
    """Returns a rotation matrix which will rotate points about the x, z, then y axis by angles in degrees."""
    angles = np.radians(xzy_axis_angles)
    cos_c, cos_a, cos_b = np.cos(angles)
    sin_c, sin_a, sin_b = np.sin(angles)

    sin_a_sin_c = sin_a * sin_c
    sin_a_cos_c = sin_a * cos_c

    rotation_matrix = np.array([
        [cos_a * cos_b, -sin_a_cos_c * cos_b + sin_b * sin_c, cos_b * sin_a_sin_c + sin_b * cos_c],
        [sin_a, cos_a * cos_c, -sin_c * cos_a],
        [-sin_b * cos_a, sin_a_cos_c * sin_b + cos_b * sin_c, -sin_b * sin_a_sin_c + cos_b * cos_c],
    ])
    return rotation_matrix


@njit
def rotation_3d_matrix_njit(xzy_axis_angles):  # pragma: no cover
    """Returns a rotation matrix which will rotate points about the x, z, then y axis by angles in degrees."""
    angles = np.radians(xzy_axis_angles)
    cos_c, cos_a, cos_b = np.cos(angles)
    sin_c, sin_a, sin_b = np.sin(angles)

    sin_a_sin_c = sin_a * sin_c
    sin_a_cos_c = sin_a * cos_c

    rotation_matrix = np.array([
        [cos_a * cos_b, -sin_a_cos_c * cos_b + sin_b * sin_c, cos_b * sin_a_sin_c + sin_b * cos_c],
        [sin_a, cos_a * cos_c, -sin_c * cos_a],
        [-sin_b * cos_a, sin_a_cos_c * sin_b + cos_b * sin_c, -sin_b * sin_a_sin_c + cos_b * cos_c],
    ])
    return rotation_matrix


def reverse_rotation_3d_matrix(xzy_axis_angles):  # pragma: no cover
    """Returns a rotation matrix which will rotate points about the y, z, then x axis by angles in degrees."""

    return rotation_3d_matrix(xzy_axis_angles).T


def rotate_vector(rotation_matrix, vector):  # pragma: no cover
    """Returns the rotated vector."""
    return np.dot(rotation_matrix, vector)


def rotate_array(rotation_matrix, a):
    """Returns a copy of array a with each vector rotated by the rotation matrix."""
    return np.matmul(rotation_matrix, a.reshape(-1, 3).T).T.reshape(a.shape)


@njit
def rotate_array_njit(rotation_matrix, a):  # pragma: no cover
    """Returns a copy of array a with each vector rotated by the rotation matrix."""
    return np.dot(rotation_matrix, a.reshape(-1, 3).T).T.reshape(a.shape)


def rotate_xyz_array_around_z_axis(a, target_xy_vector):
    """Returns a copy of array a suitable for presenting a cross-section using the resulting x,z values.

    arguments:
       a (numpy float array of shape (..., 3)): the xyz points to be rotated
       target_xy_vector (2 (or 3) floats): a vector indicating which direction in source xy space will end up
          being mapped to the positive x axis in the returned data

    returns:
       numpy float array of same shape as a

    notes:
       if the input points of a lie in a vertical plane parallel to the target xy vector, then the resulting
       points will have constant y values; in general, a full rotation of the points is applied, so resulting
       y values will indicate distance 'into the page' for non-planar or unaligned data
    """

    target_v = np.zeros(3)
    target_v[:2] = target_xy_vector[:2]
    rotation_angle = azimuth(target_v) - 90.0
    rotation_matrix = rotation_matrix_3d_axial(2, rotation_angle)  # todo: check sign of rotation angle
    return rotate_array(rotation_matrix, a)


def unit_vector_from_azimuth_and_inclination(azimuth, inclination):
    """Returns unit vector with compass bearing of azimuth and inclination off +z axis.

    note:
       assumes a left handed coordinate system with y axis north and x axis east
    """

    matrix = rotation_3d_matrix((inclination, 180.0 - azimuth, 0.0))
    return rotate_vector(matrix, np.array((0.0, 0.0, 1.0)))


def tilt_3d_matrix(azimuth, dip):
    """Returns a 3D rotation matrix for applying a dip in a certain azimuth.

    note:
       if azimuth is compass bearing in degrees, and dip is in degrees, the resulting matrix can be used
       to rotate xyz points where x values are eastings, y values are northings and z increases downwards
    """

    matrix = rotation_matrix_3d_axial(2, -azimuth)  # will yield rotation around z axis so azimuth goes north
    matrix = np.dot(matrix, rotation_matrix_3d_axial(0, dip))  # adjust for dip
    matrix = np.dot(matrix, rotation_matrix_3d_axial(2, azimuth))  # rotate azimuth back to original
    return matrix


def rotation_matrix_3d_vector(v):
    """Returns a rotation matrix which will rotate vector v to the vertical (z) axis.

    note:
       the returned matrix will map a positive z axis vector onto v
    """
    v = v / np.linalg.norm(v)
    kmat = np.array([[0, 0, -v[0]], [0, 0, -v[1]], [v[0], v[1], 0]])
    rotation_matrix = np.eye(3) + kmat + kmat.dot(kmat) * (1 / (1 + v[2]))

    rotation_matrix[:2] = -rotation_matrix[:2]
    return rotation_matrix


@njit
def rotation_matrix_3d_vector_njit(v):  # pragma: no cover
    """Returns a rotation matrix which will rotate points by inclination and azimuth of vector.

    note:
       the returned matrix will map a positive z axis vector onto v
    """
    v = v / np.linalg.norm(v)
    kmat = np.array([[0, 0, -v[0]], [0, 0, -v[1]], [v[0], v[1], 0]])
    rotation_matrix = np.eye(3) + kmat + kmat.dot(kmat) * (1 / (1 + v[2]))

    rotation_matrix[:2] = -rotation_matrix[:2]
    return rotation_matrix


def tilt_points(pivot_xyz, azimuth, dip, points):
    """Modifies array of xyz points in situ to apply dip in direction of azimuth, about pivot point."""

    matrix = tilt_3d_matrix(azimuth, dip)
    points_shape = points.shape
    points[:] -= pivot_xyz
    points[:] = np.matmul(matrix, points.reshape((-1, 3)).transpose()).transpose().reshape(points_shape)
    points[:] += pivot_xyz


def project_points_onto_plane(plane_xyz, normal_vector, points):
    """Modifies array of xyz points in situ to project onto a plane defined by a point and normal vector.

    note:
       implicit xy & z units must be the same
    """

    az = azimuth(normal_vector)
    incl = inclination(normal_vector)
    tilt_points(plane_xyz, az, -incl, points)
    points[..., 2] = plane_xyz[2]
    tilt_points(plane_xyz, az, incl, points)


def perspective_vector(xyz_box, view_axis, vanishing_distance, vector):
    """Returns a version of vector with a perspective applied."""

    mid_points = np.zeros(3)
    xyz_ranges = np.zeros(3)
    result = np.zeros(3)
    for axis in range(3):
        mid_points[axis] = 0.5 * (xyz_box[0, axis] + xyz_box[1, axis])
        xyz_ranges[axis] = xyz_box[1, axis] - xyz_box[0, axis]
    factor = 1.0 - (vector[view_axis] - xyz_box[0, view_axis]) / (vanishing_distance * (xyz_ranges[view_axis]))
    result[view_axis] = vector[view_axis]
    for axis in range(3):
        if axis == view_axis:
            continue
        result[axis] = mid_points[axis] + factor * (vector[axis] - mid_points[axis])
    return result


def determinant(a, b, c):
    """Returns the determinant of the 3 x 3 matrix comprised of the 3 vectors."""

    return (a[0] * b[1] * c[2] + a[1] * b[2] * c[0] + a[2] * b[0] * c[1] - a[2] * b[1] * c[0] - a[1] * b[0] * c[2] -
            a[0] * b[2] * c[1])


def determinant_3x3(a):
    """Returns the determinant of the 3 x 3 matrix."""

    return determinant(a[0], a[1], a[2])


def clockwise(a, b, c):
    """Returns a +ve value if 2D points a,b,c are in clockwise order, 0.0 if in line, -ve for ccw.

    note:
       assumes positive y-axis is anticlockwise from positive x-axis
    """

    return (c[0] - a[0]) * (b[1] - a[1]) - ((c[1] - a[1]) * (b[0] - a[0]))


def clockwise_triangles(p, t, projection = 'xy'):
    """Returns a numpy array of +ve values where triangle points are in clockwise order, 0.0 if in line, -ve for ccw.

    arguments:
       p (numpy float array of shape (N, 2 or 3)): points in use as vertices of triangles
       t (numpy int array of shape (M, 3)): indices into first axis of p defining the triangles
       projection (string, default 'xy'): one of 'xy', 'xz' or 'yz' being the direction of projection,
          ie. which elements of the second axis of p to use; must be 'xy' if p has shape (N, 2)

    returns:
       numpy float array of shape (M,) being the +ve or -ve values indicating clockwise or anti-clockwise
       ordering of each triangle's vertices when projected onto the specified plane and viewed in the
       direction negative to positive of the omitted axis

    note:
       assumes xyz axes are left-handed (reverse the result for a right handed system)
    """

    a0, a1 = _projected_xyz_axes(projection)

    return (((p[t[:, 2], a0] - p[t[:, 0], a0]) * (p[t[:, 1], a1] - p[t[:, 0], a1])) -
            ((p[t[:, 2], a1] - p[t[:, 0], a1]) * (p[t[:, 1], a0] - p[t[:, 0], a0])))


def in_triangle(a, b, c, d):
    """Returns True if point d lies wholly within the triangle pf ccw points a, b, c, projected onto xy plane.

    note:
       a, b & c must be sorted into anti-clockwise order before calling this function
    """

    return clockwise(a, b, d) < 0.0 and clockwise(b, c, d) < 0.0 and clockwise(c, a, d) < 0.0


def in_triangle_edged(a, b, c, d):
    """Returns True if d lies within or on the boudnary of triangle of ccw points a,b,c projected onto xy plane.

    note:
       a, b & c must be sorted into anti-clockwise order before calling this function
    """

    return clockwise(a, b, d) <= 0.0 and clockwise(b, c, d) <= 0.0 and clockwise(c, a, d) <= 0.0


def points_in_triangles(p, t, da, projection = 'xy', edged = False):
    """Returns 2D numpy bool array indicating which of points da are within which triangles.

    arguments:
       p (numpy float array of shape (N, 2 or 3)): points in use as vertices of triangles
       t (numpy int array of shape (M, 3)): indices into first axis of p defining the triangles
       da (numpy float array of shape (D, 2 or 3)): points to test for
       projection (string, default 'xy'): one of 'xy', 'xz' or 'yz' being the direction of projection,
          ie. which elements of the second axis of p  and da to use; must be 'xy' if p and da have shape (N, 2)
       edged (bool, default False): if True, points lying exactly on the edge of a triangle are included
          as being in the triangle, otherwise they are excluded

    returns:
       numpy bool array of shape (M, D) indicating which points are within which triangles

    note:
       the triangles do not need to be in a consistent clockwise or anti-clockwise order
    """

    assert p.ndim == 2 and t.ndim == 2 and da.ndim == 2 and da.shape[1] == p.shape[1]
    cwt = clockwise_triangles(p, t, projection = projection)
    a0, a1 = _projected_xyz_axes(projection)
    d_count = len(da)
    d_base = len(p)
    t_count = len(t)
    pp = np.concatenate((p, da), axis = 0)
    # build little triangles using da points and two of triangle vertices
    tp = np.empty((t_count, d_count, 3, 3), dtype = int)
    tp[:, :, :, 2] = np.arange(d_count).reshape((1, -1, 1)) + d_base
    tp[:, :, 0, 0] = np.where(cwt > 0.0, t[:, 1], t[:, 0]).reshape((-1, 1))
    tp[:, :, 0, 1] = np.where(cwt > 0.0, t[:, 0], t[:, 1]).reshape((-1, 1))
    tp[:, :, 1, 0] = np.where(cwt > 0.0, t[:, 2], t[:, 1]).reshape((-1, 1))
    tp[:, :, 1, 1] = np.where(cwt > 0.0, t[:, 1], t[:, 2]).reshape((-1, 1))
    tp[:, :, 2, 0] = np.where(cwt > 0.0, t[:, 0], t[:, 2]).reshape((-1, 1))
    tp[:, :, 2, 1] = np.where(cwt > 0.0, t[:, 2], t[:, 0]).reshape((-1, 1))
    cwtd = clockwise_triangles(pp, tp.reshape((-1, 3)), projection = projection).reshape((t_count, d_count, 3))
    if edged:
        return np.all(cwtd <= 0.0, axis = 2)
    else:
        return np.all(cwtd < 0.0, axis = 2)


@njit
def point_in_polygon(x, y, polygon):  # pragma: no cover
    """Calculates if a point in within a polygon in 2D.
    
    arguments:
        x (float): the point's x-coordinate.
        y (float): the point's y-coordinate.
        polygon (np.ndarray): array of the polygon's vertices in 2D.
    
    returns:
        inside (bool): True if point is within the polygon, False otherwise.

    note:
        the polygon is assumed closed, the closing point should not be repeated
    """
    n = len(polygon)
    inside = False
    p2x = 0.0
    p2y = 0.0
    xints = 0.0
    p1x, p1y = polygon[0]
    for i in range(n + 1):
        p2x, p2y = polygon[i % n]
        if y > min(p1y, p2y):
            if y <= max(p1y, p2y):
                if x <= max(p1x, p2x):
                    if p1y != p2y:
                        xints = (y - p1y) * (p2x - p1x) / (p2y - p1y) + p1x
                    if p1x == p2x or x <= xints:
                        inside = not inside
        p1x, p1y = p2x, p2y

    return inside


@njit
def point_in_triangle(x, y, triangle):  # pragma: no cover
    """Calculates if a point in within a triangle in 2D.

    arguments:
        x (float): the point's x-coordinate.
        y (float): the point's y-coordinate.
        triangle (np.ndarray): array of the triangles's vertices in 2D, of shape (3, 2)

    returns:
        inside (bool): True if point is within the polygon, False otherwise.
    """
    p0x = triangle[0, 0]
    p1x = triangle[1, 0]
    p2x = triangle[2, 0]
    min_p01x = min(p0x, p1x)
    if x < min(min_p01x, p2x):
        return False
    max_p01x = max(p0x, p1x)
    if x > max(max_p01x, p2x):
        return False
    p0y = triangle[0, 1]
    p1y = triangle[1, 1]
    p2y = triangle[2, 1]
    min_p01y = min(p0y, p1y)
    if y < min(min_p01y, p2y):
        return False
    max_p01y = max(p0y, p1y)
    if y > max(max_p01y, p2y):
        return False
    inside = False
    xints = 0.0
    if y > min_p01y:
        if y <= max_p01y:
            if x <= max_p01x:
                if p0x == p1x:
                    inside = True
                else:
                    if p0y == p1y:
                        inside = True
                    else:
                        xints = (y - p0y) * (p1x - p0x) / (p1y - p0y) + p0x
                        if x <= xints:
                            inside = True
    if y > min(p1y, p2y):
        if y <= max(p1y, p2y):
            if x <= max(p1x, p2x):
                if p1x == p2x:
                    inside = not inside
                else:
                    if p1y == p2y:
                        inside = not inside
                    else:
                        xints = (y - p1y) * (p2x - p1x) / (p2y - p1y) + p1x
                        if x <= xints:
                            inside = not inside
    if y > min(p2y, p0y):
        if y <= max(p2y, p0y):
            if x <= max(p2x, p0x):
                if p2x == p0x:
                    inside = not inside
                else:
                    if p2y == p0y:
                        inside = not inside
                    else:
                        xints = (y - p2y) * (p0x - p2x) / (p0y - p2y) + p2x
                        if x <= xints:
                            inside = not inside
    return inside


@njit(parallel = True)
def points_in_polygon(points: np.ndarray,
                      polygon: np.ndarray,
                      points_xlen: int,
                      polygon_num: int = 0) -> np.ndarray:  # pragma: no cover
    """Calculates which points are within a polygon in 2D.

    arguments:
        points (np.ndarray): array of shape (N, 2 or 3), of the points in 2D (xy, any z values are ignored)
        polygon (np.ndarray): list-like array of the polygon's vertices in 2D
        points_xlen (int): the original I extent of the now flattened points, use 1 if not applicable
        polygon_num (int): the polygon number, default is 0, for copying to output

    returns:
        polygon_points (np.ndarray): list-like 2D array containing only the points within the polygon,
            with each row being the polygon number (as input), points J index, and points I index

    note:
        the polygon is assumed closed, the closing point should not be repeated
    """
    polygon_points = np.full((len(points), 3), -1, dtype = np.int32)
    x = points[:, 0]
    y = points[:, 1]
    p = np.empty(len(points), dtype = np.bool_)
    j = np.empty(len(points), dtype = np.int32)
    i = np.empty(len(points), dtype = np.int32)
    for point_num in numba.prange(len(points)):
        p[point_num] = point_in_polygon(x[point_num], y[point_num], polygon)
        if p[point_num] is True:
            j[point_num], i[point_num] = divmod(point_num, points_xlen)
            polygon_points[point_num, 0] = polygon_num
            polygon_points[point_num, 1] = j[point_num]
            polygon_points[point_num, 2] = i[point_num]

    return polygon_points[polygon_points[:, 0] != -1]


@njit
def points_in_triangle(points: np.ndarray,
                       triangle: np.ndarray,
                       points_xlen: int,
                       triangle_num: int = 0) -> np.ndarray:  # pragma: no cover
    """Calculates which points are within a triangle in 2D.

    arguments:
        points (np.ndarray): array of the points in 2D.
        triangle (np.ndarray): array of the triangle's vertices in 2D, shape (3, 2).
        points_xlen (int): the number of unique x coordinates.
        triangle_num (int): the triangle number, default is 0.

    returns:
        triangle_points (np.ndarray): 2D array containing only the points within the triangle,
            with each row being the triangle number, points y index, and points x index.
    """
    triangle_points = np.full((len(points), 3), -1, dtype = np.int32)
    for point_num in numba.prange(len(points)):
        p = point_in_triangle(points[point_num, 0], points[point_num, 1], triangle)
        if p is True:
            yi, xi = divmod(point_num, points_xlen)
            triangle_points[point_num] = [triangle_num, yi, xi]

    return triangle_points[triangle_points[:, 0] != -1]


@njit
def mesh_points_in_triangle(triangle: np.ndarray,
                            points_xlen: int,
                            points_ylen: int,
                            triangle_num: int = 0) -> np.ndarray:  # pragma: no cover
    """Calculates which implicit mesh points are within a triangle in 2D for normalised triangle.

    arguments:
        triangle (np.ndarray): array of the triangle's vertices in 2D, shape (3, 2).
        points_xlen (int): the number of unique x coordinates, starting at 0.0, spacing 1.0.
        points_ylen (int): the number of unique y coordinates, starting at 0.0, spacing 1.0.
        triangle_num (int): the triangle number, default is 0.

    returns:
        triangle_points (np.ndarray): 2D array containing only the points within the triangle,
            with each row being the triangle number, points y index, and points x index.
    """
    triangle_points = np.empty((0, 3), dtype = numba.int32)
    y = 0.0
    for yi in numba.prange(points_ylen):
        x = 0.0
        for xi in numba.prange(points_xlen):
            p = point_in_triangle(x, y, triangle)
            if p is True:
                triangle_points = np.append(triangle_points,
                                            np.array([[triangle_num, yi, xi]], dtype = numba.int32),
                                            axis = 0)
            x += 1.0
        y += 1.0
    return triangle_points


@njit
def points_in_polygons(points: np.ndarray, polygons: np.ndarray, points_xlen: int) -> np.ndarray:  # pragma: no cover
    """Calculates which points are within which polygons in 2D.

    arguments:
        points (np.ndarray): array of the points in 2D.
        polygons (np.ndarray): array of each polygons' vertices in 2D.
        points_xlen (int): the number of unique x coordinates.

    returns:
        polygons_points (np.ndarray): 2D array (list-like) containing only the points within each polygon,
            with each row being the polygon number, points y index, and points x index.
    """
    polygons_points = np.empty((0, 3), dtype = numba.int32)
    for polygon_num in numba.prange(len(polygons)):
        polygon_points = points_in_polygon(points, polygons[polygon_num], points_xlen, polygon_num)
        polygons_points = np.append(polygons_points, polygon_points, axis = 0)

    return polygons_points


@njit
def points_in_triangles_njit(points: np.ndarray, triangles: np.ndarray,
                             points_xlen: int) -> np.ndarray:  # pragma: no cover
    """Calculates which points are within which triangles in 2D.

    arguments:
        points (np.ndarray): array of the points in 2D.
        triangles (np.ndarray): array of each triangles' vertices in 2D, shape (N, 3, 2).
        points_xlen (int): the number of unique x coordinates.

    returns:
        triangles_points (np.ndarray): 2D array (list-like) containing only the points within each triangle,
            with each row being the triangle number, points y index, and points x index.
    """
    triangles_points = np.empty((0, 3), dtype = numba.int32)
    for triangle_num in numba.prange(len(triangles)):
        triangle_points = points_in_triangle(points, triangles[triangle_num], points_xlen, triangle_num)
        triangles_points = np.append(triangles_points, triangle_points, axis = 0)

    return triangles_points


@njit
def meshgrid(x: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:  # pragma: no cover
    """Returns coordinate matrices from coordinate vectors x and y.

    arguments:
        x (np.ndarray): 1d array of x coordinates.
        y (np.ndarray): 1d array of y coordinates.

    returns:
        Tuple containing:

            - xx (np.ndarray): the elements of x repeated to fill the matrix along the first dimension.
            - yy (np.ndarray): the elements of y repeated to fill the matrix along the second dimension.
    """
    xx = np.empty(shape = (y.size, x.size), dtype = x.dtype)
    yy = np.empty(shape = (y.size, x.size), dtype = y.dtype)
    for i in range(y.size):
        for j in range(x.size):
            xx[i, j] = x[j]
            yy[i, j] = y[i]

    return xx, yy


def points_in_triangles_aligned(nx: int, ny: int, dx: float, dy: float, triangles: np.ndarray) -> np.ndarray:
    """Calculates which points are within which triangles in 2D for a regular mesh of aligned points.

    arguments:
        nx (int): number of points in x axis
        ny (int): number of points in y axis
        dx (float): spacing of points in x axis (first point is at half dx)
        dy (float): spacing of points in y axis (first point is at half dy)
        triangles (np.ndarray): float array of each triangles' vertices in 2D, shape (N, 3, 2).
        points_xlen (int): the number of unique x coordinates.

    returns:
        triangles_points (np.ndarray): 2D array (list-like) containing only the points within each triangle,
            with each row being the triangle number, points y index, and points x index.
    """
    triangles_points = np.empty((0, 3), dtype = np.int32)
    dx_dy = np.expand_dims(np.array([dx, dy], dtype = np.float32), axis = 0)
    # for triangle_num in numba.prange(len(triangles)):
    for triangle_num in range(len(triangles)):
        tp = (triangles[triangle_num] / dx_dy) - 0.5
        min_tpx = max(maths.ceil(min(tp[0, 0], tp[1, 0], tp[2, 0])), 0)
        max_tpx = min(maths.floor(max(tp[0, 0], tp[1, 0], tp[2, 0])), nx - 1)
        if max_tpx < min_tpx:
            continue
        min_tpy = max(maths.ceil(min(tp[0, 1], tp[1, 1], tp[2, 1])), 0)
        max_tpy = min(maths.floor(max(tp[0, 1], tp[1, 1], tp[2, 1])), ny - 1)
        if max_tpy < min_tpy:
            continue
        ntpx = max_tpx - min_tpx + 1
        ntpy = max_tpy - min_tpy + 1
        # x = np.linspace(float(min_tpx), float(max_tpx), ntpx)
        # y = np.linspace(float(min_tpy), float(max_tpy), ntpy)
        # p = np.stack(meshgrid(x, y), axis = -1).reshape((-1, 2))
        # triangle_points = points_in_triangle(p, tp, ntpx, triangle_num)
        tp[:, 0] -= float(min_tpx)
        tp[:, 1] -= float(min_tpy)
        triangle_points = mesh_points_in_triangle(tp, ntpx, ntpy, triangle_num)
        triangle_points[:, 1] += min_tpy
        triangle_points[:, 2] += min_tpx
        triangles_points = np.append(triangles_points, triangle_points, axis = 0)

    return triangles_points


@njit
def triangle_box(triangle: np.ndarray) -> Tuple[float, float, float, float]:  # pragma: no cover
    """Finds the minimum and maximum x and y values of a single traingle.

    arguments:
        triangle (np.ndarray): array of the traingle's vertices' x and y coordinates.

    returns:
        Tuple containing:

            - (float): minimum x value.
            - (float): maximum x value.
            - (float): minimum y value.
            - (float): maximum y value.
    """
    x_values = triangle[:, 0]
    y_values = triangle[:, 1]
    return min(x_values), max(x_values), min(y_values), max(y_values)


@njit
def vertical_intercept(x: float, x_values: np.ndarray, y_values: np.ndarray) -> Optional[float]:  # pragma: no cover
    """Finds the y value of a straight line between two points at a given x.
    
    If the x value given is not within the x values of the points, returns None.

    arguments:
        x (float): x value at which to determine the y value.
        x_values (np.ndarray): the x coordinates of point 1 and point 2.
        y_values (np.ndarray): the y coordinates of point 1 and point 2.

    returns:
        y (Optional[float]): y value of the straight line between point 1 and point 2,
            evaluated at x. If x is outside the x_values range, y is None.
    """
    y = None
    if x >= min(x_values) and x <= max(x_values):
        if x_values[0] == x_values[1]:
            y = y_values[0]
        else:
            m = (y_values[1] - y_values[0]) / (x_values[1] - x_values[0])
            c = y_values[1] - m * x_values[1]
            y = m * x + c
    return y


@njit
def points_in_triangles_aligned_optimised(nx: int, ny: int, dx: float, dy: float,
                                          triangles: np.ndarray) -> np.ndarray:  # pragma: no cover
    """Calculates which points are within which triangles in 2D for a regular mesh of aligned points.

    arguments:
        nx (int): number of points in x axis
        ny (int): number of points in y axis
        dx (float): spacing of points in x axis (first point is at half dx)
        dy (float): spacing of points in y axis (first point is at half dy)
        triangles (np.ndarray): float array of each triangles' vertices in 2D, shape (N, 3, 2)

    returns:
        triangles_points (np.ndarray): 2D array (list-like) containing only the points within each triangle,
            with each row being the triangle number, points y index, and points x index
    """
    grid_x = dx * (0.5 + np.arange(nx).astype(np.float64))
    grid_y = dy * (0.5 + np.arange(ny).astype(np.float64))
    triangles_points_list = []
    for triangle_num in range(len(triangles)):
        triangle = triangles[triangle_num]
        min_x, max_x, min_y, max_y = triangle_box(triangle)
        x_values = grid_x[np.logical_and(grid_x >= min_x, grid_x < max_x)]
        y_values = grid_y[np.logical_and(grid_y >= min_y, grid_y < max_y)]
        for x in x_values:
            ys = []
            ys.append(vertical_intercept(x, triangle[1:, 0], triangle[1:, 1]))
            ys.append(vertical_intercept(x, triangle[:2, 0], triangle[:2, 1]))
            ys.append(vertical_intercept(x, triangle[::2, 0], triangle[::2, 1]))
            ys = [u for u in ys if u is not None]
            valid_y = y_values[np.logical_and(y_values >= min(ys), y_values <= max(ys))]
            x_idx = int(x / dx)
            triangles_points_list.extend([[triangle_num, int(y / dy), x_idx] for y in valid_y])

    if len(triangles_points_list) == 0:
        triangles_points = np.empty((0, 3), dtype = np.int32)
        return triangles_points

    triangles_points = np.array(triangles_points_list, dtype = np.int32)
    return triangles_points


def triangle_normal_vector(p3):
    """For a triangle in 3D space, defined by 3 vertex points, returns a unit vector normal to the plane of the triangle.

    note:
        resulting vector implicitly assumes that xy & z units are the same; if this is not the case, adjust vector
        afterwards as required
    """

    # todo: handle degenerate triangles
    return unit_vector(cross_product(p3[0] - p3[1], p3[0] - p3[2]))


@njit
def triangle_normal_vector_numba(points):  # pragma: no cover
    """For a triangle in 3D space, defined by 3 vertex points, returns a unit vector normal to the plane of the triangle.

    note:
        resulting vector implicitly assumes that xy & z units are the same; if this is not the case, adjust vector
        afterwards as required
    """
    v = np.cross(points[0] - points[1], points[0] - points[2])
    return v / np.linalg.norm(v)


def in_circumcircle(a, b, c, d):
    """Returns True if point d lies within the circumcircle pf ccw points a, b, c, projected onto xy plane.

    note:
       a, b & c must be sorted into anti-clockwise order before calling this function
    """

    m = np.empty((3, 3))
    m[0, :2] = a[:2] - d[:2]
    m[1, :2] = b[:2] - d[:2]
    m[2, :2] = c[:2] - d[:2]
    m[:, 2] = (m[:, 0] * m[:, 0]) + (m[:, 1] * m[:, 1])
    return determinant_3x3(m) > 0.0


def point_distance_to_line_2d(p, l1, l2):
    """Ignoring any z values, returns the xy distance of point p from line passing through l1 and l2."""

    if np.all(l2[:2] == l1[:2]):
        return naive_2d_length(p[:2] - l1[:2])
    return (abs(p[0] * (l1[1] - l2[1]) + l1[0] * (l2[1] - p[1]) + l2[0] * (p[1] - l1[1])) /
            naive_2d_length(l2[:2] - l1[:2]))


def point_distance_to_line_segment_2d(p, l1, l2):
    """Ignoring any z values, returns the xy distance of point p from line segment between l1 and l2."""

    if is_obtuse_2d(l1, p, l2):
        return naive_2d_length(p[:2] - l1[:2])
    elif is_obtuse_2d(l2, p, l1):
        return naive_2d_length(p[:2] - l2[:2])
    else:
        return point_distance_to_line_2d(p, l1, l2)


def is_obtuse_2d(p, p1, p2):
    """Returns True if the angle at point p subtended by points p1 and p2, in xy plane, is greater than 90 degrees; else False."""

    return np.dot((p1[:2] - p[:2]), (p2[:2] - p[:2])) < 0.0


def isclose(a, b, tolerance = 1.0e-6):
    """Returns True if the two points are extremely close to one another (ie.

    the same point).
    """

    #   return np.all(np.isclose(a, b, atol = tolerance))
    # cheap and cheerful alternative to thorough numpy version commented out above
    return np.max(np.abs(a - b)) <= tolerance


def is_close(a, b, tolerance = 1.0e-6):
    """Returns True if the two points are extremely close to one another (ie.

    the same point).
    """

    return isclose(a, b, tolerance = tolerance)


def point_distance_sqr_to_points_projected(p, points, projection):
    """Returns an array of projected distances squared between p and points; projection is 'xy', 'xz' or 'yz'."""

    if projection == 'xy':
        d = points[..., :2] - p[:2]
    elif projection == 'xz':
        d = points[..., 0:3:2] - p[0:3:2]
    elif projection == 'yz':
        d = points[..., 1:] - p[1:]
    else:
        raise ValueError("projection must be 'xy', 'xz' or 'yz'")
    return np.sum(d * d, axis = -1)


def nearest_point_projected(p, points, projection):
    """Returns the index into points array closest to point p; projection is 'xy', 'xz' or 'yz'."""

    # note: in the case of equidistant points, the index of the 'first' point is returned

    d2 = point_distance_sqr_to_points_projected(p, points, projection)
    return np.unravel_index(np.nanargmin(d2), d2.shape)


def area_of_triangle(a, b, c):
    """Returns the area of the triangle defined by three vertices."""

    # uses Heron's formula
    la = naive_length(a - b)
    lb = naive_length(b - c)
    lc = naive_length(c - a)
    s = 0.5 * (la + lb + lc)
    return maths.sqrt(s * (s - la) * (s - lb) * (s - lc))


def area_of_triangles(p, t, xy_projection = False):
    """Returns numpy array of areas of triangles, optionally when projected onto xy plane."""

    # uses Heron's formula
    pt = p[t]
    if xy_projection:
        la = naive_2d_lengths(pt[:, 0, :] - pt[:, 1, :])
        lb = naive_2d_lengths(pt[:, 1, :] - pt[:, 2, :])
        lc = naive_2d_lengths(pt[:, 2, :] - pt[:, 0, :])
    else:
        la = naive_lengths(pt[:, 0, :] - pt[:, 1, :])
        lb = naive_lengths(pt[:, 1, :] - pt[:, 2, :])
        lc = naive_lengths(pt[:, 2, :] - pt[:, 0, :])
    s = 0.5 * (la + lb + lc)
    return np.sqrt(s * (s - la) * (s - lb) * (s - lc))


def clockwise_sorted_indices(p, b):
    """Returns a clockwise sorted numpy list of indices b into the points p.

    note:
       this function is designed for preparing a list of points defining a convex polygon when projected in
       the xy plane, starting from a subset of the unsorted points; more specifically, it assumes that the
       mean of p (over axis 0) lies within the polygon and the clockwise ordering is relative to that mean point
    """

    # note: this function currently assumes that the mean of points bp lies within the hull of bp
    # and that the points form a convex polygon from the perspective of the mean point
    assert p.ndim == 2 and len(p) >= 3
    centre = np.mean(p, axis = 0)
    hull_list = []  # list of azimuths and indices into p (axis 0)
    for i in b:
        azi = azimuth(p[i] - centre)
        hull_list.append((azi, i))
    return np.array([i for (_, i) in sorted(hull_list)], dtype = int)


def xy_sorted(p, axis = None):
    """Returns copy of points p sorted according to x or y (whichever has greater range).

    arguments:
        p (numpy float array of shape (..., 2) or (..., 3)): points to be sorted
        axis (int, optional): 0 for x sort; 1 for y sort; None for whichever has greater range

    returns:
        p', axis where p' is a list-like (2D) version of p, sorted by either x or y and axis is 0
        if the sort was by x, 1 if it were by y

    note:
       returned array is always 2D, ie. list of points
    """
    assert p.ndim >= 2 and p.shape[-1] >= 2
    p = p.reshape((-1, p.shape[-1]))
    if axis is None:
        xy_range = np.nanmax(p, axis = 0) - np.nanmin(p, axis = 0)
        axis = (1 if xy_range[1] > xy_range[0] else 0)
    spi = np.argsort(p[:, axis])
    return p[spi], axis


@njit
def xy_sorted_njit(p, axis = -1):  # pragma: no cover
    """Returns copy of points p sorted according to x or y (whichever has greater range)."""
    assert p.ndim >= 2 and p.shape[-1] >= 2
    p = p.reshape((-1, p.shape[-1]))
    if axis == -1:
        xy_range = _nanmax(p) - _nanmin(p)
        axis = (1 if xy_range[1] > xy_range[0] else 0)
    points = p[:, axis]
    spi = np.argsort(points)
    return p[spi], axis


@njit
def _nanmax(array):  # pragma: no cover
    """Numba implementation of np.nanmax with axis = 0."""
    len0 = array.shape[1]
    max_array = np.empty(len0)
    for col in range(len0):
        col_array = array[:, col]
        max_val = col_array[0]
        for val in col_array:
            if val > max_val and not np.isnan(val):
                max_val = val
        max_array[col] = max_val
    return max_array


@njit
def _nanmin(array):  # pragma: no cover
    """Numba implementation of np.nanmin with axis = 0."""
    len0 = array.shape[1]
    min_array = np.empty(len0)
    for col in range(len0):
        col_array = array[:, col]
        min_val = col_array[0]
        for val in col_array:
            if val < min_val and not np.isnan(val):
                min_val = val
        min_array[col] = min_val
    return min_array


def _projected_xyz_axes(projection):
    assert projection in ['xy', 'xz', 'yz'], f'invalid projection {projection}'
    a0 = 'xyz'.index(projection[0])
    a1 = 'xyz'.index(projection[1])
    return a0, a1


# end of vector_utilities module



"""Class for RESQML Geobody Boudary Interpretation organizational objects."""

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.organize
import resqpy.organize.genetic_boundary_feature as gbf
import resqpy.organize._utils as ou
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns


class GeobodyBoundaryInterpretation(BaseResqpy):
    """Class for RESQML Geobody Boundary Interpretation organizational objects."""

    resqml_type = 'GeobodyBoundaryInterpretation'
    valid_domains = ('depth', 'time', 'mixed')
    valid_boundary_relations = ('conformable', 'unconformable below and above', 'unconformable above',
                                'unconformable below')

    def __init__(self,
                 parent_model,
                 uuid = None,
                 title = None,
                 genetic_boundary_feature = None,
                 domain = 'depth',
                 boundary_relation_list = None,
                 originator = None,
                 extra_metadata = None):
        """ Instantiate a GeobodyBoundaryInterpretation object.

        arguments:
            parent_model(model.Model): Model to which the feature belongs
            uuid(UUID, Optional): The UUID of an existing Geobody Boundary Interpretation object; if present,
                all the other optional arguments are ignored
            title(str, Optional): Citation title when creating a new object
            genetic_boundary_feature(GeneticBoundaryFeature,Optional): Interpreted feature when creating a new object
            domain(str,Optional): One of ('depth', 'time', 'mixed') when creating a new object
            boundary_relation_list(list,Optional): Set of ('conformable', 'unconformable below and above', 'unconformable above',
                'unconformable below') when creating a new object
            extra_metadata(dict,Optional): Extra metadata items added when creating a new object
        """

        self.domain = domain
        self.boundary_relation_list = None if not boundary_relation_list else boundary_relation_list.copy()
        self.genetic_boundary_feature = genetic_boundary_feature  # InterpretedFeature RESQML field, when not loading from xml
        self.feature_root = None if self.genetic_boundary_feature is None else self.genetic_boundary_feature.root
        if (not title) and self.genetic_boundary_feature is not None:
            title = self.genetic_boundary_feature.feature_name
        self.has_occurred_during = (None, None)
        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

    def _load_from_xml(self):
        root_node = self.root
        self.domain = rqet.find_tag_text(root_node, 'Domain')
        interp_feature_ref_node = rqet.find_tag(root_node, 'InterpretedFeature')
        assert interp_feature_ref_node is not None
        self.feature_root = self.model.referenced_node(interp_feature_ref_node)
        if self.feature_root is not None:
            self.genetic_boundary_feature = gbf.GeneticBoundaryFeature(self.model,
                                                                       kind = 'geobody boundary',
                                                                       uuid = self.feature_root.attrib['uuid'],
                                                                       feature_name = self.model.title_for_root(
                                                                           self.feature_root))
        self.has_occurred_during = ou.extract_has_occurred_during(root_node)
        br_node_list = rqet.list_of_tag(root_node, 'BoundaryRelation')
        if br_node_list is not None and len(br_node_list) > 0:
            self.boundary_relation_list = []
            for br_node in br_node_list:
                self.boundary_relation_list.append(br_node.text)

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this interpretation is essentially the same as the other; otherwise False."""

        if other is None or not isinstance(other, GeobodyBoundaryInterpretation):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if self.genetic_boundary_feature is not None:
            if not self.genetic_boundary_feature.is_equivalent(other.genetic_boundary_feature,
                                                               check_extra_metadata = check_extra_metadata):
                return False
        elif other.genetic_boundary_feature is not None:
            return False
        if self.root is not None and other.root is not None:
            if rqet.citation_title_for_node(self.root) != rqet.citation_title_for_node(other.root):
                return False
        elif self.root is not None or other.root is not None:
            return False
        if ((self.domain != other.domain) or
                not ou.equivalent_chrono_pairs(self.has_occurred_during, other.has_occurred_during)):
            return False
        if check_extra_metadata and not ou.equivalent_extra_metadata(self, other):
            return False
        if not self.boundary_relation_list and not other.boundary_relation_list:
            return True
        if not self.boundary_relation_list or not other.boundary_relation_list:
            return False
        return set(self.boundary_relation_list) == set(other.boundary_relation_list)

    def create_xml(self,
                   genetic_boundary_feature_root = None,
                   add_as_part = True,
                   add_relationships = True,
                   originator = None,
                   title_suffix = None,
                   reuse = True):
        """Create a organisational xml node from a geobody boundary interpretation object."""

        if not self.title:
            self.title = self.genetic_boundary_feature.feature_name
        if title_suffix:
            self.title += ' ' + title_suffix

        if reuse and self.try_reuse():
            return self.root
        gbi = super().create_xml(add_as_part = False, originator = originator)

        if self.genetic_boundary_feature is not None:
            gbf_root = self.genetic_boundary_feature.root
            if gbf_root is not None:
                if genetic_boundary_feature_root is None:
                    genetic_boundary_feature_root = gbf_root
                else:
                    assert gbf_root is genetic_boundary_feature_root, 'genetic boundary feature mismatch'
        else:
            if genetic_boundary_feature_root is None:
                genetic_boundary_feature_root = self.feature_root
            assert genetic_boundary_feature_root is not None
            self.genetic_boundary_feature = gbf.GeneticBoundaryFeature(
                self.model, uuid = genetic_boundary_feature_root.attrib['uuid'])
        self.feature_root = genetic_boundary_feature_root

        assert self.domain in self.valid_domains, 'illegal domain value for geobody boundary interpretation'
        dom_node = rqet.SubElement(gbi, ns['resqml2'] + 'Domain')
        dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
        dom_node.text = self.domain

        ou.create_xml_has_occurred_during(self.model, gbi, self.has_occurred_during)

        if self.boundary_relation_list is not None:
            for boundary_relation in self.boundary_relation_list:
                assert boundary_relation in self.valid_boundary_relations
                br_node = rqet.SubElement(gbi, ns['resqml2'] + 'BoundaryRelation')
                br_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'BoundaryRelation')
                br_node.text = str(boundary_relation)

        self.model.create_ref_node('InterpretedFeature',
                                   self.model.title_for_root(genetic_boundary_feature_root),
                                   genetic_boundary_feature_root.attrib['uuid'],
                                   content_type = 'obj_GeneticBoundaryFeature',
                                   root = gbi)

        if add_as_part:
            self.model.add_part('obj_GeobodyBoundaryInterpretation', self.uuid, gbi)
            if add_relationships:
                self.model.create_reciprocal_relationship(gbi, 'destinationObject', genetic_boundary_feature_root,
                                                          'sourceObject')

        return gbi



"""Module containing common methods for properties"""

# Nexus is a trademark of Halliburton

import logging

log = logging.getLogger(__name__)

import warnings
import numpy as np

import resqpy.property as rqp
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.weights_and_measures as bwam

# the following resqml property kinds and facet types are 'known about' by this module in relation to nexus
# other property kinds should be handled okay but without any special treatment
# see property_kind_and_facet_from_keyword() for simulator keyword to property kind and facet mapping

supported_property_kind_list = [
    'continuous', 'discrete', 'categorical', 'code', 'index', 'depth', 'rock volume', 'pore volume', 'volume',
    'thickness', 'length', 'cell length', 'area', 'net to gross ratio', 'porosity', 'permeability thickness',
    'permeability length', 'permeability rock', 'rock permeability', 'fluid volume', 'transmissibility', 'pressure',
    'saturation', 'solution gas-oil ratio', 'vapor oil-gas ratio', 'property multiplier', 'thermodynamic temperature'
]

supported_local_property_kind_list = [
    'active', 'transmissibility multiplier', 'fault transmissibility', 'mat transmissibility'
]

supported_facet_type_list = ['direction', 'netgross', 'what']

# current implementation limits a property to having at most one facet; resqml standard allows for many
# if a property kind does not appear in the following dictionary, facet_type and facet are expected to be None
# mapping from property kind to (facet_type, list of possible facet values), only applicable when indexable element is 'cells'
# use of the following in code is DEPRECATED; it remains here to give a hint of facet types and facets in use

expected_facet_type_dict = {
    'depth': ('what', ['cell centre', 'cell top']),  # made up
    'rock volume': ('netgross', ['net', 'gross']),  # resqml standard
    'thickness': ('netgross', ['net', 'gross']),
    'length': ('direction', ['X', 'Y', 'Z']),  # facet values made up
    'cell length': ('direction', ['I', 'J', 'K']),  # facet values made up
    'permeability rock': ('direction', ['I', 'J', 'K']),  # todo: allow IJ and IJK
    'rock permeability': ('direction', ['I', 'J', 'K']),  # todo: allow IJ and IJK
    'transmissibility': ('direction', ['I', 'J', 'K']),
    'transmissibility multiplier': (
        'direction',  # local property kind
        ['I', 'J', 'K']),
    'fault transmissibility': ('direction', ['I', 'J']),  # local property kind; K faces also permitted
    'mat transmissibility': ('direction', ['K']),  # local property kind; I & J faces also permitted
    'saturation': (
        'what',
        [
            'water',
            'oil',
            'gas',  # made up but probably good
            'water minimum',
            'gas minimum',
            'oil minimum',  # made up for end-points
            'water residual',
            'gas residual',
            'oil residual',
            'water residual to oil',
            'gas residual to oil'
        ]),
    'fluid volume': (
        'what',
        [
            'water',
            'oil',
            'gas',  # made up but probably good
            'water (mobile)',
            'oil (mobile)',
            'gas (mobile)'
        ]),  # made up
    'property multiplier': (
        'what',
        [
            'rock volume',  # made up; todo: add rock permeability?
            'pore volume',
            'transmissibility'
        ]),
    'code': ('what', ['inactive', 'active']),  # NB: user defined keywords also accepted
    'index': ('what', ['uid', 'pixel map'])
}


def same_property_kind(pk_a, pk_b):
    """Returns True if the two property kinds are the same, or pseudonyms."""

    if pk_a is None or pk_b is None:
        return False
    if pk_a == pk_b:
        return True
    if pk_a in ['permeability rock', 'rock permeability'] and pk_b in ['permeability rock', 'rock permeability']:
        return True
    return False


def property_kind_and_facet_from_keyword(keyword):
    """If keyword is recognised, returns equivalent resqml PropertyKind and Facet info.

    argument:
       keyword (string): Nexus grid property keyword

    returns:
       (property_kind, facet_type, facet) as defined in resqml standard; some or all may be None

    note:
       this function may now return the local property kind 'transmissibility multiplier'; calling code must ensure that
       the local property kind object is created if not already present
    """
    # note: this code doesn't cater for a property having more than one facet, eg. direction and phase
    # note: 'what' facet_type for made-up uses might be better changed to the generic 'qualifier' facet type

    property_kind = None
    facet_type = None
    facet = None
    lk = keyword.lower()

    if lk in ['bv', 'brv', 'pv', 'pvr', 'porv', 'netv', 'nrv']:
        property_kind, facet_type, facet = _pkf_from_keyword_rock_volume(lk)
    elif lk in ['ntg', 'netgrs']:
        property_kind = 'net to gross ratio'  # net-to-gross
    elif lk in ['por', 'poro', 'porosity']:
        property_kind = 'porosity'  # porosity
    elif lk == 'kh':
        property_kind = 'permeability thickness'  # K.H (not horizontal permeability)
    elif lk in ['p', 'pressure']:
        property_kind = 'pressure'  # pressure; todo: phase pressures
    elif lk == 'rs':
        property_kind = 'solution gas-oil ratio'
    elif lk == 'rv':
        property_kind = 'vapor oil-gas ratio'
    elif lk in ['temp', 'temperature']:
        property_kind = 'thermodynamic temperature'
    elif lk in ['sw', 'so', 'sg', 'satw', 'sato', 'satg', 'soil']:  # saturations
        property_kind, facet_type, facet = _pkf_from_keyword_saturation(lk)
    elif lk in ['swl', 'swr', 'sgl', 'sgr', 'swro', 'sgro', 'sor', 'swu', 'sgu']:  # nexus saturation end points
        property_kind, facet_type, facet = _pkf_from_keyword_saturation_end(lk)
    elif lk in ['dzc', 'dzn', 'dz', 'dznet']:
        property_kind, facet_type, facet = _pkf_from_keyword_thickness(lk)
    elif lk in ['mdep', 'depth', 'tops', 'mids']:
        property_kind, facet_type, facet = _pkf_from_keyword_depth(lk)
    elif lk == 'uid':
        property_kind = 'index'
        facet_type = 'what'
        facet = 'uid'
    elif lk in ['wip', 'oip', 'gip', 'mobw', 'mobo', 'mobg', 'ocip']:  # todo: check these, especially ocip
        property_kind, facet_type, facet = _pkf_from_keyword_fluid_volume(lk)
    elif lk in ['tmx', 'tmy', 'tmz', 'tmflx', 'tmfly', 'tmflz', 'multx', 'multy', 'multz']:
        property_kind, facet_type, facet = _pkf_from_keyword_transmissibility_multiplier(lk)
    elif lk in ['multbv', 'multpv']:
        property_kind, facet_type, facet = _pkf_from_keyword_property_multiplier(lk)
    elif lk in ['dad', 'kid', 'unpack', 'deadcell', 'inactive']:
        property_kind, facet_type, facet = _pkf_from_keyword_inactive(lk)
    elif len(lk) >= 2 and lk[0] == 'd' and lk[1] in 'xyz':
        property_kind, facet_type, facet = _pkf_from_keyword_length(lk)
    elif lk[:4] == 'perm' or (len(lk) == 2 and lk[0] == 'k'):  # permeability
        property_kind = 'permeability rock'
        (facet_type, facet) = _facet_info_for_dir_ch(lk[-1])
    elif lk[:5] == 'trans' or (len(lk) == 2 and lk[0] == 't'):  # transmissibility (for unit viscosity)
        if 'mult' in lk:
            property_kind = 'transmissibility multiplier'
        else:
            property_kind = 'transmissibility'
        if lk != 'transmissibility':
            (facet_type, facet) = _facet_info_for_dir_ch(lk[-1])
    # elif lk == 'sal':    # todo: salinity; local property kind needed; various units possible in Nexus
    elif lk == 'livecell' or lk.startswith('act'):
        property_kind = 'active'  # local property kind, see RESQML (2.0.1) usage guide, section 11.17
    elif len(lk) <= 8 and (lk.startswith('reg') or lk.startswith('creg')):
        property_kind = 'region initialization'  # local property kind, see RESQML (2.0.1) usage guide, section 11.18
    elif lk in ['iequil', 'ipvt', 'irelpm', 'irock', 'iregion', 'iwater', 'ireg']:
        property_kind = 'region initialization'  # local property kind, see RESQML (2.0.1) usage guide, section 11.18
        facet_type = 'what'
        facet = lk[1:]
        if facet == 'equil':
            facet = 'equilibration'
        elif facet == 'relpm':
            facet = 'relative permeability'
    return property_kind, facet_type, facet


def _pkf_from_keyword_length(lk):
    if lk in ['dxc', 'dyc', 'dx', 'dy']:
        property_kind = 'cell length'
        (facet_type, facet) = _facet_info_for_dir_ch(lk[1])
    else:
        property_kind = 'length'
        facet_type = 'direction'
        facet = lk[1].upper()  # keep as 'X', 'Y' or 'Z'
    return property_kind, facet_type, facet


def _pkf_from_keyword_rock_volume(lk):
    if lk in ['bv', 'brv']:  # bulk rock volume
        return 'rock volume', 'netgross', 'gross'  # todo: check this is the facet in xsd
    elif lk in ['pv', 'pvr', 'porv']:
        return 'pore volume', None, None
    elif lk in ['netv', 'nrv']:  # net volume
        return 'rock volume', 'netgross', 'net'
    else:
        return None, None, None  # should never come to this


def _pkf_from_keyword_fluid_volume(lk):
    property_kind = 'fluid volume'
    facet_type = 'what'  # todo: check use of 'what' for phase
    facet = ''
    if lk in ['wip', 'mobw']:
        facet = 'water'  # todo: add another facet indicating mobile volume
    elif lk in ['oip', 'mobo']:
        facet = 'oil'
    elif lk in ['gip', 'mobg']:
        facet = 'gas'
    elif lk == 'ocip':
        facet = 'oil condensate'  # todo: this seems unlikely: check
    if lk[:3] == 'mob':
        facet += ' (mobile)'
    return property_kind, facet_type, facet


def _facet_info_for_dir_ch(dir_ch):
    facet_type = None
    facet = None
    if dir_ch in ['i', 'j', 'k', 'x', 'y', 'z']:
        facet_type = 'direction'
        if dir_ch in ['i', 'x']:
            facet = 'I'
        elif dir_ch in ['j', 'y']:
            facet = 'J'
        else:
            facet = 'K'
        # NB: resqml also allows for combinations, eg. IJ, IJK
    return facet_type, facet


def _pkf_from_keyword_property_multiplier(lk):
    property_kind = 'property multiplier'
    facet_type = 'what'  # here 'what' facet indicates property affected
    if lk == 'multbv':
        facet = 'rock volume'  # NB: making this up as I go along
    elif lk == 'multpv':
        facet = 'pore volume'
    return property_kind, facet_type, facet


def _pkf_from_keyword_inactive(lk):
    property_kind = 'code'
    facet_type = 'what'
    # todo: kid can only be used as an inactive cell indication for the root grid
    if lk in ['kid', 'deadcell', 'inactive']:
        facet = 'inactive'  # standize on 'inactive' to indicate use as mask
    else:
        facet = lk  # note: use deadcell or unpack for inactive, if nothing better?
    return property_kind, facet_type, facet


def _pkf_from_keyword_saturation_end(lk):
    property_kind = 'saturation'
    facet_type = 'what'  # note: use of 'what' for phase is a guess
    facet = ''
    if lk[1] == 'w':
        facet = 'water'
    elif lk[1] == 'g':
        facet = 'gas'
    elif lk[1] == 'o':
        facet = 'oil'
    if lk[-1] == 'l':
        facet += ' minimum'
    elif lk[-1] == 'u':
        facet += ' maximum'
    elif lk[2:] == 'ro':
        facet += ' residual to oil'
    elif lk[-1] == 'r':
        facet += ' residual'
    else:
        assert False, 'problem deciphering saturation end point keyword: ' + lk
    return property_kind, facet_type, facet


def _pkf_from_keyword_saturation(lk):
    property_kind = 'saturation'
    facet_type = 'what'  # todo: check use of 'what' for phase
    if lk in ['sw', 'satw', 'swat']:
        facet = 'water'
    elif lk in ['so', 'sato', 'soil']:
        facet = 'oil'
    elif lk in ['sg', 'satg', 'sgas']:
        facet = 'gas'
    return property_kind, facet_type, facet


def _pkf_from_keyword_thickness(lk):
    property_kind = 'thickness'  # or should these keywords use cell length in K direction?
    facet_type = 'netgross'
    if lk.startswith('dzn'):
        facet = 'net'
    else:
        facet = 'gross'
    return property_kind, facet_type, facet


def _pkf_from_keyword_depth(lk):
    property_kind = 'depth'  # depth (nexus) and tops mean top depth
    facet_type = 'what'  # this might need to be something different
    if lk in ['mdep', 'mids']:
        facet = 'cell centre'
    else:
        facet = 'cell top'
    return property_kind, facet_type, facet


def _pkf_from_keyword_transmissibility_multiplier(lk):
    property_kind = 'transmissibility multiplier'  # NB: resqpy local property kind
    facet_type = 'direction'
    _, facet = _facet_info_for_dir_ch(lk[-1])
    return property_kind, facet_type, facet


def infer_property_kind(name, unit):
    """Guess a valid property kind."""

    # Currently unit is ignored

    valid_kinds = bwam.valid_property_kinds()

    if name in valid_kinds:
        kind = name
    else:
        # TODO: use an appropriate default
        kind = 'Unknown'

    # TODO: determine facet_type and facet somehow
    facet_type = None
    facet = None

    return kind, facet_type, facet


def _crs_m_or_ft(crs_node):  # NB. models not-so-rarely use metres for xy and feet for z
    if crs_node is None:
        return None
    xy_units = rqet.find_tag(crs_node, 'ProjectedUom').text.lower()
    z_units = rqet.find_tag(crs_node, 'VerticalUom').text.lower()
    if xy_units == 'm' and z_units == 'm':
        return 'm'
    if xy_units == 'ft' and z_units == 'ft':
        return 'ft'
    if xy_units == 'cm' and z_units == 'cm':
        return 'cm'
    return None


def guess_uom(property_kind, minimum, maximum, support, facet_type = None, facet = None):
    """Returns a guess at the units of measure for the given kind of property.

    arguments:
       property_kind (string): a valid resqml property kind, lowercase
       minimum: the minimum value in the data for which the units are being guessed
       maximum: the maximum value in the data for which the units are being guessed
       support: the grid.Grid or well.WellboreFrame object which the property data relates to
       facet_type (string, optional): a valid resqml facet type, lowercase, one of:
               'direction', 'what', 'netgross', 'qualifier', 'conditions', 'statistics'
       facet: (string, present if facet_type is present): the value relating to the facet_type,
                eg. 'I' for direction, or 'oil' for 'what'

    returns:
       a valid resqml unit of measure (uom) for the property_kind, or None

    notes:
       this function is tailored towards Nexus unit systems;
       the resqml standard allows a property to have any number of facets; however,
       this module currently only supports zero or one facet per property
    """
    crs_node, from_crs = _guess_uom_get_crs_info(support)

    if property_kind in ['rock volume', 'pore volume', 'volume', 'fluid volume']:
        return _guess_uom_volume(property_kind, from_crs, facet_type, facet)
    if property_kind == 'depth':
        return _guess_uom_depth(crs_node)
    if property_kind == 'cell length':  # todo: pass in direction facet to pick up xy_units or z_units
        return from_crs
    if property_kind in ['net to gross ratio', 'porosity', 'saturation']:
        return _guess_uom_ntg_por_sat(maximum, from_crs)
    if property_kind == 'permeability rock' or property_kind == 'rock permeability':
        return 'mD'
    if property_kind in ['permeability thickness', 'permeability length']:
        return _guess_uom_permeability_thickness(property_kind, crs_node, from_crs)
    if property_kind.endswith('transmissibility'):
        return _guess_uom_transmissibility(from_crs)
    if property_kind == 'pressure':
        return _guess_uom_pressure(from_crs, maximum)
    if property_kind in ['solution gas-oil ratio', 'vapor oil-gas ratio']:
        return _guess_uom_gor_ogr(property_kind, from_crs)
    if property_kind.endswith('multiplier'):
        return 'Euc'
    # todo: 'degC' or 'degF' for thermodynamic temperature
    return None


def _guess_uom_get_crs_info(support):
    if support is None or not hasattr(support, 'extract_crs_uuid'):
        crs_node = None
    else:
        crs_node = support.model.root_for_uuid(support.extract_crs_uuid())
    return crs_node, _crs_m_or_ft(crs_node)


def _guess_uom_depth(crs_node):
    if crs_node is None:
        return None
    return rqet.find_tag(crs_node, 'VerticalUom').text.lower()


def _guess_uom_volume(property_kind, from_crs, facet_type, facet):
    if property_kind == 'fluid volume':
        if from_crs == 'm':
            return 'm3'
        if from_crs == 'ft':
            if facet_type == 'what' and facet == 'gas':
                return '1000 ft3'  # todo: check units output by nexus for GIP
            else:
                return 'bbl'  # todo: check whether nexus uses 10^3 or 10^6 units
        if from_crs == 'cm':
            return 'cm3'
        return None
    if from_crs is None:
        return None
    if from_crs == 'ft' and property_kind == 'pore volume':
        return 'bbl'  # seems to be Nexus 'ENGLISH' uom for pv out
    return from_crs + '3'  # ie. m3 or ft3 or cm3


def _guess_uom_permeability_thickness(property_kind, crs_node, from_crs):
    z_units = rqet.find_tag(crs_node, 'VerticalUom').text.lower()
    if z_units in ['m', 'cm']:  # note: mD.cm is not a valid RESQML uom
        return 'mD.m'
    if z_units == 'ft':
        return 'mD.ft'
    return None


def _guess_uom_ntg_por_sat(maximum, from_crs):
    if maximum is not None and str(maximum) != 'unknown':
        max_real = float(maximum)
        if max_real > 1.0 and max_real <= 100.0:
            return '%'
        if max_real < 0.0 or max_real > 1.0:
            return None
    if from_crs == 'm':
        return 'm3/m3'
    if from_crs == 'ft':
        return 'ft3/ft3'
    if from_crs == 'cm':
        return 'cm3/cm3'
    return 'Euc'


def _guess_uom_transmissibility(from_crs):
    # note: RESQML QuantityClass only includes a unit-viscosity VolumePerTimePerPressureUom
    if from_crs == 'm':
        return 'm3.cP/(kPa.d)'  # NB: might actually be m3/(psi.d) or m3/(bar.d)
    if from_crs == 'ft':
        return 'bbl.cP/(psi.d)'  # gamble on barrels per day per psi; could be ft3/(psi.d)
    if from_crs == 'cm':
        return 'cm3.cP/(psi.d)'  # complete guess, though transmissibility probably not used in lab scale models
    return None


def _guess_uom_pressure(from_crs, maximum):
    if from_crs == 'm':
        return 'kPa'  # NB: might actually be psi or bar
    if from_crs in ['ft', 'cm']:  # note: Nexus uses psi for lab pressure units
        return 'psi'
    if maximum is not None:
        max_real = float(maximum)
        if max_real == 0.0:
            return None
        if max_real > 10000.0:
            return 'kPa'
        if max_real < 500.0:
            return 'bar'
        if max_real < 5000.0:
            return 'psi'
    return None


def _guess_uom_gor_ogr(property_kind, from_crs):
    if from_crs == 'm':
        return 'm3/m3'  # NB: might actually be psi or bar
    if from_crs == 'ft':
        if 'solution' in property_kind:
            return '1000 ft3/bbl'
        else:
            return '0.001 bbl/ft3'
    if from_crs == 'cm':
        return 'cm3/cm3'
    return None


def selective_version_of_collection(collection,
                                    realization = None,
                                    support_uuid = None,
                                    uuid = None,
                                    continuous = None,
                                    points = None,
                                    count = None,
                                    indexable = None,
                                    property_kind = None,
                                    facet_type = None,
                                    facet = None,
                                    citation_title = None,
                                    time_series_uuid = None,
                                    time_index = None,
                                    uom = None,
                                    string_lookup_uuid = None,
                                    categorical = None,
                                    title = None,
                                    title_mode = None,
                                    related_uuid = None,
                                    const_value = None,
                                    extra = None):
    """Returns a new PropertyCollection with those parts which match all arguments that are not None.

    arguments:
       collection (PropertyCollection): an existing collection from which a subset will be returned as a new object
       realization (int, optional): realization number to filter on
       support_uuid (UUID or str, optional): UUID of supporting representation, to filter on
       uuid (UUID or str, optional): a property uuid to select a singleton property from the collection
       continuous (bool, optional): if True, continuous properties are selected; if False, discrete and categorical
       points (bool, optional): if True, points properties are selected; if False, they are excluded
       count (int, optional): a count value to filter on
       indexable (str, optional): indexable elements flavour to filter on
       property_kind (str, optional): property kind to filter on (commonly used)
       facet_type (str, optional): a facet_type to filter on (must be present for property to be selected)
       facet (str, optional): a facet value to filter on (if used, facet_type should be specified)
       citation_title (str, optional): citation to title to filter on; see also title_mode argument
       time_series_uuid (UUID or str, optional): UUID of a TimeSeries to filter on
       time_index (int, optional): a time series time index to filter on
       uom (str, optional): unit of measure to filter on
       string_lookup_uuid (UUID or str, optional): UUID of a string lookup table to filter on
       categorical (bool, optional): if True, only categorical properties are selected; if False they are excluded
       title (str, optional): synonymous with citation_title argument
       title_mode (str, optional): if present, one of 'is', 'starts', 'ends', 'contains', 'is not',
           'does not start', 'does not end', 'does not contain'; None is the same as 'is'; will default to 'is'
           if not specified and title or citation_title argument is present
       related_uuid (UUID or str, optional): only properties with direct relationship to this uuid are selected
       const_value (float or int, optional): only properties flagged as constant, with given value, are selected
       extra (dict, optional): only properties where the extra metadata includes all the items of this dict are selected

    returns:
       a new PropertyCollection containing those properties which match the filter parameters that are not None

    notes:
       the existing collection might often be the 'main' collection holding all the properties
       for a supporting representation (eg. grid, blocked well or wellbore frame);
       for each of the filtering arguments: if None, then all members of collection pass this filter;
       if not None then only those members with the given value pass this filter;
       special values: '*' any non-None value passes; 'none' only None passes;
       citation_title (or its synonym title) uses string filtering in association with title_mode argument;
       finally, the filters for all the attributes must be passed for a given member
       to be included in the returned collection; title is a synonym for the citation_title argument;
       related_uuid will pass if any relationship exists ('hard' or 'soft');
       the categorical boolean argument can be used to select only categorical (or non-categorical) properties,
       even though this is not explicitly held as a field in the internal dictionary
    """

    assert collection is not None
    view = rqp.PropertyCollection()
    if support_uuid is not None:
        view.set_support(support_uuid = support_uuid, model = collection.model)
    if realization is not None:
        view.set_realization(realization)
    if citation_title is None:
        citation_title = title
    view.inherit_parts_selectively_from_other_collection(collection,
                                                         realization = realization,
                                                         support_uuid = support_uuid,
                                                         uuid = uuid,
                                                         continuous = continuous,
                                                         points = points,
                                                         count = count,
                                                         indexable = indexable,
                                                         property_kind = property_kind,
                                                         facet_type = facet_type,
                                                         facet = facet,
                                                         citation_title = citation_title,
                                                         citation_title_match_mode = title_mode,
                                                         time_series_uuid = time_series_uuid,
                                                         time_index = time_index,
                                                         uom = uom,
                                                         string_lookup_uuid = string_lookup_uuid,
                                                         categorical = categorical,
                                                         related_uuid = related_uuid,
                                                         const_value = const_value,
                                                         extra = extra)
    return view


def property_over_time_series_from_collection(collection, example_part):
    """Returns a new PropertyCollection with parts like the example part, over all indices in its time series.

    arguments:
       collection: an existing PropertyCollection from which a subset will be returned as a new object;
                   the existing collection might often be the 'main' collection holding all the properties
                   for a grid
       example_part (string): the part name of an example member of collection (which has an associated time_series)

    returns:
       a new PropertyCollection containing those memners of collection which have the same property kind
       (and facet, if any) as the example part and which have the same associated time series
    """

    assert collection is not None and example_part is not None
    assert collection.part_in_collection(example_part)
    view = rqp.PropertyCollection()
    if collection.support_uuid is not None:
        view.set_support(support_uuid = collection.support_uuid, model = collection.model)
    if collection.realization is not None:
        view.set_realization(collection.realization)
    view.inherit_similar_parts_for_time_series_from_other_collection(collection, example_part)
    return view


def property_collection_for_keyword(collection, keyword):
    """Returns a new PropertyCollection with parts that match the property kind and facet deduced for the keyword.

    arguments:
       collection: an existing PropertyCollection from which a subset will be returned as a new object;
                   the existing collection might often be the 'main' collection holding all the properties
                   for a supporting representation (grid or wellbore frame)
       keyword (string): a simulator keyword for which the property kind (and facet, if any) can be deduced

    returns:
       a new PropertyCollection containing those memners of collection which have the property kind
       (and facet, if any) as that deduced for the keyword

    note:
       this function is particularly relevant to grid property collections for simulation models;
       the handling of simulator keywords in this module is based on the main grid property keywords
       for Nexus; if the resqml dataset was generated from simulator data using this module then
       the result of this function should be reliable; resqml data sets from other sources might use facets
       if a different way, leading to an omission in the results of this function
    """

    assert collection is not None and keyword
    (property_kind, facet_type, facet) = property_kind_and_facet_from_keyword(keyword)
    if property_kind is None:
        log.warning('failed to deduce property kind for keyword: ' + keyword)
        return None
    return selective_version_of_collection(collection,
                                           property_kind = property_kind,
                                           facet_type = facet_type,
                                           facet = facet)


def reformat_column_edges_to_resqml_format(array):
    """Converts an array of shape (nj,ni,2,2) to shape (nj,ni,4) in RESQML edge ordering."""
    newarray = np.empty((array.shape[0], array.shape[1], 4), dtype = array.dtype)
    newarray[:, :, 0] = array[:, :, 1, 0]
    newarray[:, :, 1] = array[:, :, 0, 1]
    newarray[:, :, 2] = array[:, :, 1, 1]
    newarray[:, :, 3] = array[:, :, 0, 0]
    return newarray


def reformat_column_edges_from_resqml_format(array):
    """Converts an array of shape (nj,ni,4) in RESQML edge ordering to shape (nj,ni,2,2)"""
    newarray = np.empty((array.shape[0], array.shape[1], 2, 2), dtype = array.dtype)
    newarray[:, :, 0, 0] = array[:, :, 3]
    newarray[:, :, 0, 1] = array[:, :, 1]
    newarray[:, :, 1, 0] = array[:, :, 0]
    newarray[:, :, 1, 1] = array[:, :, 2]
    return newarray


# 'private' functions returning attribute name for cached version of property array
# I find the leading underscore so ugly, I can't bring myself to use it for 'private' functions, even though many people do


def _cache_name_for_uuid(uuid):
    """Returns the attribute name used for the cached copy of the property array for the given uuid.

    :meta private:
    """

    return 'c_' + bu.string_from_uuid(uuid)


def _cache_name(part):
    """Returns the attribute name used for the cached copy of the property array for the given part.

    :meta private:
    """

    if part is None:
        return None
    uuid = rqet.uuid_in_part_name(part)
    if uuid is None:
        return None
    return _cache_name_for_uuid(uuid)


def dtype_flavour(continuous, use_32_bit):
    """Returns the numpy elemental data type depending on the two boolean flags.

    :meta private:
    """

    if continuous:
        if use_32_bit:
            dtype = np.float32
        else:
            dtype = np.float64
    else:
        if use_32_bit:
            dtype = np.int32
        else:
            dtype = np.int64
    return dtype


def return_cell_indices(i, cell_indices):
    """Returns the i'th entry in the cell_indices array, or NaN if i has null value of -1."""

    if i == -1:
        return np.nan
    else:
        return cell_indices[i]


def write_hdf5_and_create_xml_for_active_property(model,
                                                  active_property_array,
                                                  support_uuid,
                                                  title = 'ACTIVE',
                                                  realization = None,
                                                  time_series_uuid = None,
                                                  time_index = None):
    """Writes hdf5 data and creates xml for an active cell property; returns uuid."""

    active = rqp.Property.from_array(parent_model = model,
                                     cached_array = active_property_array,
                                     source_info = None,
                                     keyword = title,
                                     support_uuid = support_uuid,
                                     property_kind = 'active',
                                     local_property_kind_uuid = None,
                                     indexable_element = 'cells',
                                     discrete = True,
                                     time_series_uuid = time_series_uuid,
                                     time_index = time_index,
                                     realization = realization,
                                     find_local_property_kind = True)
    return active.uuid


def check_and_warn_property_kind(pk, activity):
    """Check property kind and warn if one of three main abstract kinds."""
    if pk in ['continuous', 'discrete', 'categorical']:
        warnings.warn(
            f"abstract property kind '{pk}', whilst {activity}, will be or have been replaced by local property kind")
        # raise ValueError


def property_parts(model,
                   obj_type,
                   parts_list = None,
                   property_kind = None,
                   facet_type = None,
                   facet = None,
                   related_uuid = None):
    """Returns list of property parts from model matching filters."""

    def facet_match(root, facet_type, facet):
        if not facet_type and not facet:
            return True
        assert facet_type and facet is not None
        facets = rqet.list_of_tag(root, 'Facet')
        if facet_type == 'none' and facets:
            return False
        for f_node in facets:
            if rqet.find_tag_text(f_node, 'Facet') == facet_type:
                if facet == 'none':
                    return False
                if facet == '*':
                    return True
                return rqet.find_tag_text(f_node, 'Value') == str(facet)
        if facet == 'none':
            return True
        return False

    if not obj_type.endswith('Property'):
        obj_type += 'Property'
    assert obj_type in ['ContinuousProperty', 'DiscreteProperty', 'CategoricalProperty', 'PointsProperty']
    parts = model.parts(parts_list = parts_list, obj_type = obj_type, related_uuid = related_uuid)
    if property_kind:
        pk_parts = []
        for part in parts:
            root = model.root_for_part(part)
            node = rqet.find_nested_tags(root, ['PropertyKind', 'Kind'])
            if node is None:
                # following relies on title in reference xml matching that of the local proparty kind
                node = rqet.find_nested_tags(root, ['PropertyKind', 'LocalPropertyKind', 'Title'])
                assert node is not None
            if node.text != property_kind and (property_kind not in ['rock permeability', 'permeability rock'] or
                                               node.text not in ['rock permeability', 'permeability rock']):
                continue
            if facet_match(root, facet_type, facet):
                pk_parts.append(part)
        parts = pk_parts
    elif facet_type and facet is not None:
        f_parts = []
        for part in parts:
            root = model.root_for_part(part)
            if facet_match(root, facet_type, facet):
                pk_parts.append(part)
    return parts


def property_part(model,
                  obj_type,
                  parts_list = None,
                  property_kind = None,
                  facet_type = None,
                  facet = None,
                  related_uuid = None):
    """Returns individual property part from model matching filters."""
    parts = property_parts(model,
                           obj_type,
                           parts_list = parts_list,
                           property_kind = property_kind,
                           facet_type = facet_type,
                           facet = facet,
                           related_uuid = related_uuid)
    if parts is None or len(parts) == 0:
        return None
    assert len(parts) == 1, 'more than one property part matches criteria'
    return parts[0]



"""Containing resqml propertykind class"""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns


class PropertyKind(BaseResqpy):
    """Class catering for RESQML bespoke PropertyKind objects."""

    resqml_type = "PropertyKind"

    def __init__(self,
                 parent_model,
                 uuid = None,
                 title = None,
                 is_abstract = False,
                 example_uom = None,
                 naming_system = 'urn:resqml:bp.com:resqpy',
                 parent_property_kind = 'continuous',
                 extra_metadata = None,
                 originator = None):
        """Initialise a new bespoke property kind."""

        self.is_abstract = is_abstract
        self.naming_system = naming_system
        self.example_uom = example_uom
        self.parent_kind = parent_property_kind
        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

    def _load_from_xml(self):
        root_node = self.root
        self.is_abstract = rqet.find_tag_bool(root_node, 'IsAbstract')
        self.naming_system = rqet.find_tag_text(root_node, 'NamingSystem')
        self.example_uom = rqet.find_tag_text(root_node, 'RepresentativeUom')
        ppk_node = rqet.find_tag(root_node, 'ParentPropertyKind')
        assert ppk_node is not None
        ppk_kind_node = rqet.find_tag(ppk_node, 'Kind')
        assert ppk_kind_node is not None, 'only standard property kinds supported as parent kind'
        self.parent_kind = ppk_kind_node.text

    def is_equivalent(self, other_pk, check_extra_metadata = True):
        """Returns True if this property kind is essentially the same as the other; False otherwise."""

        if other_pk is None:
            return False
        if self is other_pk:
            return True
        if bu.matching_uuids(self.uuid, other_pk.uuid):
            return True
        if (self.parent_kind != other_pk.parent_kind or self.title != other_pk.title or
                self.is_abstract != other_pk.is_abstract or self.naming_system != other_pk.naming_system):
            return False
        if (self.example_uom and other_pk.example_uom) and self.example_uom != other_pk.example_uom:
            return False
        if check_extra_metadata:
            if (self.extra_metadata or other_pk.extra_metadata) and self.extra_metadata != other_pk.extra_metadata:
                return False
        return True

    def create_xml(self, add_as_part = True, originator = None, reuse = True):
        """Create xml for this bespoke property kind."""

        if reuse and self.try_reuse():
            return self.root  # check for reusable (equivalent) object

        pk = super().create_xml(add_as_part = False, originator = originator)

        ns_node = rqet.SubElement(pk, ns['resqml2'] + 'NamingSystem')
        ns_node.set(ns['xsi'] + 'type', ns['xsd'] + 'anyURI')
        ns_node.text = str(self.naming_system)

        ia_node = rqet.SubElement(pk, ns['resqml2'] + 'IsAbstract')
        ia_node.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
        ia_node.text = str(self.is_abstract).lower()

        # note: schema definition requires this field, even for discrete property kinds
        uom = self.example_uom
        if uom is None:
            uom = 'Euc'
        ru_node = rqet.SubElement(pk, ns['resqml2'] + 'RepresentativeUom')
        ru_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ResqmlUom')
        ru_node.text = str(uom)

        ppk_node = rqet.SubElement(pk, ns['resqml2'] + 'ParentPropertyKind')
        ppk_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'StandardPropertyKind')
        ppk_node.text = rqet.null_xml_text

        ppk_kind_node = rqet.SubElement(ppk_node, ns['resqml2'] + 'Kind')
        ppk_kind_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ResqmlPropertyKind')
        ppk_kind_node.text = str(self.parent_kind)

        if add_as_part:
            self.model.add_part('obj_PropertyKind', self.uuid, pk)
            # no relationships at present, if local parent property kinds were to be supported then a rel. is needed there

        return pk


def create_transmisibility_multiplier_property_kind(model):
    """Create a local property kind 'transmisibility multiplier' for a given model.

    argument:
       model: resqml model object

    returns:
       property kind uuid
    """
    log.debug("Making a new property kind 'Transmissibility multiplier'")
    tmult_kind = PropertyKind(parent_model = model,
                              title = 'transmissibility multiplier',
                              parent_property_kind = 'continuous')
    tmult_kind.create_xml()
    tmult_kind_uuid = tmult_kind.uuid
    model.store_epc()
    return tmult_kind_uuid


def establish_zone_property_kind(model):
    """Returns zone local property kind object, creating the xml and adding as part if not found in model."""

    zone_pk_uuid = model.uuid(obj_type = 'LocalPropertyKind', title = 'zone')
    if zone_pk_uuid is None:
        zone_pk = PropertyKind(model, title = 'zone', parent_property_kind = 'discrete')
        zone_pk.create_xml()
    else:
        zone_pk = PropertyKind(model, uuid = zone_pk_uuid)
    return zone_pk



"""_polyline_set.py: Resqml polyline set module."""

import logging

log = logging.getLogger(__name__)

import os
import numpy as np

import resqpy.crs as rcrs
import resqpy.lines
import resqpy.property as rqp
import resqpy.olio.simple_lines as rsl
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vu
import resqpy.olio.write_hdf5 as rwh5
import resqpy.olio.xml_et as rqet
import resqpy.lines._common as rql_c
from resqpy.olio.xml_namespaces import curly_namespace as ns


class PolylineSet(rql_c._BasePolyline):
    """Class for RESQML polyline set representation."""

    resqml_type = 'PolylineSetRepresentation'

    def __init__(self,
                 parent_model,
                 uuid = None,
                 polylines = None,
                 irap_file = None,
                 charisma_file = None,
                 crs_uuid = None,
                 title = None,
                 originator = None,
                 extra_metadata = None):
        """Initialises a new PolylineSet object.

        arguments:
            parent_model (model.Model object): the model which the new PolylineSetRepresentation belongs to
            uuid (uuid.UUID, optional): the uuid of an existing RESQML PolylineSetRepresentation object from
                which to initialise this resqpy PolylineSet
            polylines (optional): list of polyline objects from which to build the polylineset
            irap_file (str, optional): the name of a file in irap format from which to import the polyline set
            charisma_file (str, optional): the name of a file in charisma format from which to import the polyline set
            crs_uuid (UUID, optional): required if loading from an IRAP or Charisma file
            title (str, optional): the citation title to use for a new polyline set;
                ignored if uuid is not None
            originator (str, optional): the name of the person creating the polyline set, defaults to login id;
                ignored if uuid is not None
            extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the polyline set;
                ignored if uuid is not None

        returns:
            the newly instantiated PolylineSet object

        note:
            this initialiser does not perform any conversion between CRSes; if loading from a list of Polylines,
            they must all share a common CRS (which must also match crs_uuid if supplied)

        :meta common:
        """

        self.model = parent_model
        self.coordinates = None
        self.count_perpol = None
        self.polys = []
        self.rep_int_root = None
        self.save_polys = False
        self.boolnotconstant = None
        self.boolvalue = None
        self.closed_array = None
        self.crs_uuid = crs_uuid
        self.indices = None
        self.property_collection = None

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.root is not None:
            return

        if polylines is not None:  # Create from list of polylines
            crs_list = [] if crs_uuid is None else [bu.uuid_from_string(crs_uuid)]
            for poly in polylines:
                crs_list.append(poly.crs_uuid)
            crs_set = set(crs_list)
            # following assumes consistent use of UUID or str
            assert len(crs_set) == 1, 'More than one CRS found in input polylines for polyline set'
            for crs_uuid in crs_set:
                self.crs_uuid = crs_uuid
                if self.crs_uuid is not None:
                    break
            self.polys = polylines
            # Setting the title of the first polyline given as the PolylineSet title
            if not self.title:
                if len(polylines) > 1:
                    self.title = f"{polylines[0].title} + {len(polylines)-1} polylines"
                else:
                    self.title = polylines[0].title
            self.combine_polylines(polylines)
        elif irap_file is not None:  # Create from an input IRAP file
            if crs_uuid is None:
                log.warning(f'applying generic model CRS when loading polylines from IRAP file: {irap_file}')
            self._set_from_irap(irap_file)

        elif charisma_file is not None:
            if crs_uuid is None:
                log.warning(f'applying generic model CRS when loading polylines from Charisma file: {charisma_file}')
            self._set_from_charisma(charisma_file)

    def _load_from_xml(self):

        assert self.root is not None  # polyline set xml node specified
        root = self.root

        self.rep_int_root = self.model.referenced_node(rqet.find_tag(root, 'RepresentedInterpretation'))
        self.closed_array = np.empty((0,), dtype = bool)

        for patch_node in rqet.list_of_tag(root, 'LinePatch'):  # Loop over all LinePatches - likely just the one
            assert patch_node is not None  # Required field

            geometry_node = rqet.find_tag(patch_node, 'Geometry')
            assert geometry_node is not None  # Required field

            crs_root = self.model.referenced_node(rqet.find_tag(geometry_node, 'LocalCrs'))
            assert crs_root is not None  # Required field
            uuid = rqet.uuid_for_part_root(crs_root)
            assert uuid is not None  # Required field
            if self.crs_uuid is not None:
                assert bu.matching_uuids(self.crs_uuid, uuid), 'mixed CRS uuids in use in PolylineSet'
            self.crs_uuid = uuid

            closed_node = rqet.find_tag(patch_node, 'ClosedPolylines')
            assert closed_node is not None  # Required field
            # The ClosedPolylines could be a BooleanConstantArray, or a BooleanArrayFromIndexArray
            closed_array = self.get_bool_array(closed_node)

            count_node = rqet.find_tag(patch_node, 'NodeCountPerPolyline')
            rql_c.load_hdf5_array(self, count_node, 'count_perpol', tag = 'Values', dtype = 'int')

            points_node = rqet.find_tag(geometry_node, 'Points')
            assert points_node is not None  # Required field
            rql_c.load_hdf5_array(self, points_node, 'coordinates', tag = 'Coordinates')

            # Check that the number of bools aligns with the number of count_perpoly
            # Check that the total of the count_perpoly aligns with the number of coordinates
            assert len(self.count_perpol) == len(closed_array)
            assert np.sum(self.count_perpol) == len(self.coordinates)

            subpolys = self.convert_to_polylines(closed_array, self.count_perpol, self.coordinates, self.crs_uuid,
                                                 self.rep_int_root)
            # Check we have the right number of polygons
            assert len(subpolys) == len(self.count_perpol)

            # Remove duplicate coordinates and count arrays (exist in polylines now)
            # delattr(self,'coordinates')
            # delattr(self,'count_perpol')

            self.polys.extend(subpolys)
            self.closed_array = np.concatenate((self.closed_array, closed_array))

        self.bool_array_format()

    def _set_from_irap(self, irap_file):
        inpoints = rsl.read_lines(irap_file)
        self.count_perpol = []
        closed_array = []
        if not self.title:
            self.title = os.path.basename(irap_file).split(".")[0]
        for i, poly in enumerate(inpoints):
            if len(poly) > 1:  # Polylines must have at least 2 points
                self.count_perpol.append(len(poly))
                if vu.isclose(poly[0], poly[-1]):
                    closed_array.append(True)
                else:
                    closed_array.append(False)
                if i == 0:
                    self.coordinates = poly
                else:
                    self.coordinates = np.concatenate((self.coordinates, poly))
        self.count_perpol = np.array(self.count_perpol)
        if self.crs_uuid is None:  # If no crs_uuid is provided, assume the main model crs is valid
            self.crs_uuid = self.model.crs_uuid
        self.polys = self.convert_to_polylines(closed_array, self.count_perpol, self.coordinates, self.crs_uuid,
                                               self.rep_int_root)

    def _set_from_charisma(self, charisma_file):
        with open(charisma_file) as f:
            inpoints = f.readlines()
        self.count_perpol = []
        closed_array = []
        if not self.title:
            self.title = os.path.basename(charisma_file).split(".")[0]
        for i, line in enumerate(inpoints):
            line = line.split()
            coord_entry = np.array([[float(line[3]), float(line[4]), float(line[5])]])
            if i == 0:
                self.coordinates = coord_entry
                stick = line[7]
                count = 1
            else:
                self.coordinates = np.concatenate((self.coordinates, coord_entry))
                count += 1
                if stick != line[7] or i == len(inpoints) - 1:
                    if count <= 2 and stick != line[7]:  # Line has fewer than 2 points
                        log.warning(f"Polylines must contain at least 2 points - ignoring point {self.coordinates[-2]}")
                        self.coordinates = np.delete(self.coordinates, -2, 0)  # Remove the second to last entry
                    else:
                        self.count_perpol.append(count - 1)
                        closed_array.append(False)
                    count = 1
                    stick = line[7]
        self.count_perpol = np.array(self.count_perpol)
        if self.crs_uuid is None:  # If no crs_uuid is provided, assume the main model crs is valid
            self.crs_uuid = self.model.crs_uuid
        self.polys = self.convert_to_polylines(closed_array, self.count_perpol, self.coordinates, self.crs_uuid,
                                               self.rep_int_root)

    def poly_index_containing_point_in_xy(self, p, mode = 'crossing'):
        """Returns the index of the first (closed) polyline containing point p in the xy plane, or None.

        :meta common:
        """

        assert mode in ['crossing', 'winding'], 'unrecognised mode when looking for polygon containing point'

        for i, poly in enumerate(self.polys):
            if not poly.isclosed:
                continue
            if poly.point_is_inside_xy(p, mode = mode):
                return i
        return None

    def create_xml(self,
                   ext_uuid = None,
                   add_as_part = True,
                   add_relationships = True,
                   title = None,
                   originator = None,
                   save_polylines = False):
        """Create xml from polylineset.

        args:
            save_polylines: If true, polylines are also saved individually

        :meta common:
        """

        if ext_uuid is None:
            ext_uuid = self.model.h5_uuid()

        self.save_polys = save_polylines

        if title:
            self.title = title
        if not self.title:
            self.title = 'polyline set'

        if self.save_polys:
            for poly in self.polys:
                poly.create_xml(ext_uuid, add_relationships = add_relationships, originator = originator)

        polyset = super().create_xml(add_as_part = False, originator = originator)

        if self.rep_int_root is not None:
            rep_int = self.rep_int_root
            if "FaultInterpretation" in str(rqet.content_type(rep_int)):
                content_type = 'obj_FaultInterpretation'
            else:
                content_type = 'obj_HorizonInterpretation'
            self.model.create_ref_node('RepresentedInterpretation',
                                       rqet.citation_title_for_node(rep_int),
                                       rep_int.attrib['uuid'],
                                       content_type = content_type,
                                       root = polyset)

        # We convert all Polylines to the CRS of the first Polyline in the set, so set this as crs_uuid

        patch = rqet.SubElement(polyset, ns['resqml2'] + 'LinePatch')
        patch.set(ns['xsi'] + 'type', ns['resqml2'] + 'PolylineSetPatch')
        patch.text = '\n'

        pindex = rqet.SubElement(patch, ns['resqml2'] + 'PatchIndex')
        pindex.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
        pindex.text = '0'

        if self.boolnotconstant:

            # We have mixed data - use a BooleanArrayFromIndexArray
            closed = rqet.SubElement(patch, ns['resqml2'] + 'ClosedPolylines')
            closed.set(ns['xsi'] + 'type', ns['xsd'] + 'BooleanArrayFromIndexArray')
            closed.text = '\n'

            bool_val = rqet.SubElement(closed, ns['resqml2'] + 'IndexIsTrue')
            bool_val.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
            bool_val.text = str(not self.boolvalue).lower()

            count = rqet.SubElement(closed, ns['resqml2'] + 'Count')
            count.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            count.text = str(len(self.count_perpol))

            ind_val = rqet.SubElement(closed, ns['resqml2'] + 'Indices')
            ind_val.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
            ind_val.text = '\n'

            iv_null = rqet.SubElement(ind_val, ns['resqml2'] + 'NullValue')
            iv_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
            iv_null.text = '-1'

            iv_values = rqet.SubElement(ind_val, ns['resqml2'] + 'Values')
            iv_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
            iv_values.text = '\n'

            self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'indices_patch0', root = iv_values)

        else:

            # All bools are the same - use a BooleanConstantArray
            closed = rqet.SubElement(patch, ns['resqml2'] + 'ClosedPolylines')
            closed.set(ns['xsi'] + 'type', ns['resqml2'] + 'BooleanConstantArray')
            closed.text = '\n'

            bool_val = rqet.SubElement(closed, ns['resqml2'] + 'Value')
            bool_val.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
            bool_val.text = str(self.boolvalue).lower()

            count = rqet.SubElement(closed, ns['resqml2'] + 'Count')
            count.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            count.text = str(len(self.count_perpol))

        count_pp = rqet.SubElement(patch, ns['resqml2'] + 'NodeCountPerPolyline')
        count_pp.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
        count_pp.text = '\n'

        null = rqet.SubElement(count_pp, ns['resqml2'] + 'NullValue')
        null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
        null.text = '0'

        count_val = rqet.SubElement(count_pp, ns['resqml2'] + 'Values')
        count_val.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        count_val.text = '\n'

        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'NodeCountPerPolyline_patch0', root = count_val)

        geom = rqet.SubElement(patch, ns['resqml2'] + 'Geometry')
        geom.set(ns['xsi'] + 'type', ns['resqml2'] + 'PointGeometry')
        geom.text = '\n'

        self.model.create_crs_reference(crs_uuid = self.crs_uuid, root = geom)

        points = rqet.SubElement(geom, ns['resqml2'] + 'Points')
        points.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dHdf5Array')
        points.text = '\n'

        coords = rqet.SubElement(points, ns['resqml2'] + 'Coordinates')
        coords.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        coords.text = '\n'

        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'points_patch0', root = coords)

        if add_as_part:
            self.model.add_part('obj_PolylineSetRepresentation', self.uuid, polyset)
            if add_relationships:
                crs_root = self.model.root_for_uuid(self.crs_uuid)
                self.model.create_reciprocal_relationship(polyset, 'destinationObject', crs_root, 'sourceObject')
                if self.rep_int_root is not None:  # Optional
                    self.model.create_reciprocal_relationship(polyset, 'destinationObject', self.rep_int_root,
                                                              'sourceObject')
                if self.save_polys:
                    for poly in self.polys:
                        self.model.create_reciprocal_relationship(polyset, 'destinationObject', poly.root_node,
                                                                  'sourceObject')

                ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
                ext_node = self.model.root_for_part(ext_part)
                self.model.create_reciprocal_relationship(polyset, 'mlToExternalPartProxy', ext_node,
                                                          'externalPartProxyToMl')

        return polyset

    def write_hdf5(self, file_name = None, mode = 'a', save_polylines = False):
        """Create or append the coordinates, counts and indices hdf5 arrays to hdf5 file.

        :meta common:
        """

        if self.uuid is None:
            self.uuid = bu.new_uuid()
        self.combine_polylines(self.polys)
        self.bool_array_format()
        self.save_polys = save_polylines
        if self.save_polys:
            for poly in self.polys:
                poly.write_hdf5(file_name)

        h5_reg = rwh5.H5Register(self.model)
        h5_reg.register_dataset(self.uuid, 'points_patch0', self.coordinates)
        h5_reg.register_dataset(self.uuid, 'NodeCountPerPolyline_patch0', self.count_perpol.astype(np.int32))
        if self.boolnotconstant:
            h5_reg.register_dataset(self.uuid, 'indices_patch0', np.array(self.indices, dtype = np.int32))
        h5_reg.write(file_name, mode = mode)

    def get_bool_array(self, closed_node):
        # TODO: check for other permissible forms of the abstract boolean array
        """Returns a boolean array using details in the xml node location.

        arguments:
            closed_node (xml node): the node under which the boolean array information sits

        returns:
            1D numpy bool array set True for those pplylines within the set which are marked as closed
        """
        # if type of boolean array is BooleanConstantArray, uses the array value and count to generate
        # the array; if type of boolean array is BooleanArrayFromIndexArray, finds the "other" value
        # bool and indices of the "other" values, and inserts these into an array opposite to the main bool
        flavour = rqet.node_type(closed_node)
        if flavour == 'BooleanConstantArray':
            count = rqet.find_tag_int(closed_node, 'Count')
            value = rqet.bool_from_text(rqet.node_text(rqet.find_tag(closed_node, 'Value')))
            return np.full((count,), value, dtype = bool)
        elif flavour == 'BooleanArrayFromIndexArray':
            count = rqet.find_tag_int(closed_node, 'Count')
            indices_node = rqet.find_tag(closed_node, 'Indices')
            assert indices_node is not None
            indices_arr = rql_c.load_hdf5_array(self, indices_node, 'indices_arr', tag = 'Values', dtype = 'int')
            istrue = rqet.bool_from_text(rqet.node_text(rqet.find_tag(closed_node, 'IndexIsTrue')))
            out = np.full((count,), not istrue, dtype = bool)
            out[indices_arr] = istrue
            return out
        raise ValueError(f'unrecognised closed array xml node type: {flavour}')

    def convert_to_polylines(self,
                             closed_array = None,
                             count_perpol = None,
                             coordinates = None,
                             crs_uuid = None,
                             rep_int_root = None):
        """Returns a list of Polylines objects from a PolylineSet.

        note:
            all arguments are optional and by default the data will be taken from self

        args:
            closed_array: array containing a bool for each polygon in if it is open (False) or closed (True)
            count_perpol: array containing a list of polygon "lengths" for each polygon
            coordinates: array containing coordinates for all the polygons
            crs_uuid: crs_uuid for polylineset
            rep_int_root: represented interpretation root (optional)

        returns:
            list of polyline objects

        :meta common:
        """

        from resqpy.lines._polyline import Polyline

        if count_perpol is None:
            count_perpol = self.count_perpol
        if closed_array is None:
            closed_array = np.zeros(len(count_perpol), dtype = bool)
            closed_node = rqet.find_nested_tags(self.root, ['LinePatch', 'ClosedPolylines'])
            if closed_node is not None:
                closed_array[:] = self.get_bool_array(closed_node)
        if coordinates is None:
            coordinates = self.coordinates
        if crs_uuid is None:
            crs_uuid = self.crs_uuid
        if rep_int_root is None:
            rep_int_root = self.rep_int_root
        polys = []
        count = 0
        for i in range(len(count_perpol)):
            if i != len(count_perpol) - 1:
                subset = coordinates[count:int(count_perpol[i]) + count].copy()
            else:
                subset = coordinates[count:int(count_perpol[i]) + count + 1].copy()
            if vu.isclose(subset[0], subset[-1]):
                isclosed = True
            else:
                isclosed = closed_array[i]
            count += int(count_perpol[i])
            subtitle = f"{self.title} {i+1}"
            polys.append(
                Polyline(self.model,
                         is_closed = isclosed,
                         set_coord = subset,
                         set_crs = crs_uuid,
                         title = subtitle,
                         rep_int_root = rep_int_root))

        return polys

    def combine_polylines(self, polylines):
        """Combines the isclosed boolean array, coordinates and count data for a list of polyline objects.

        args:
            polylines: list of polyline objects
        """

        self.count_perpol = []
        self.closed_array = []

        for poly in polylines:
            if poly == polylines[0]:
                master_crs = rcrs.Crs(self.model, uuid = poly.crs_uuid)
                self.crs_uuid = poly.crs_uuid
                self.coordinates = poly.coordinates.copy()
            else:
                curr_crs = rcrs.Crs(self.model, uuid = poly.crs_uuid)
                if not curr_crs.is_equivalent(master_crs):
                    shifted = curr_crs.convert_array_to(master_crs, poly.coordinates)
                    self.coordinates = np.concatenate((self.coordinates, shifted))
                else:
                    self.coordinates = np.concatenate((self.coordinates, poly.coordinates))

            self.closed_array.append(poly.isclosed)
            self.count_perpol.append(int(len(poly.coordinates)))

        self.count_perpol = np.array(self.count_perpol)

        assert len(self.closed_array) == len(self.count_perpol)
        assert np.sum(self.count_perpol) == len(self.coordinates)

        self.polys = polylines

    def bool_array_format(self):
        """Determines an appropriate output boolean array format depending on the closed_array bools.

        self.boolnotconstant - set to True if all are not open or all closed
        self.boolvalue - value of isclosed for all polylines, or for the majority of polylines if mixed
        self.indices - array of indices where the values are not self.boolvalue, if the polylines are mixed
        """

        assert self.closed_array is not None
        self.indices = []
        self.boolnotconstant = False
        if all(self.closed_array):
            self.boolvalue = True
        elif not any(self.closed_array):
            self.boolvalue = False
        else:
            if np.count_nonzero(self.closed_array) > (len(self.closed_array) // 2):
                self.boolvalue = True
                for i, val in enumerate(self.closed_array):
                    if not val:
                        self.indices.append(i)
            else:
                self.boolvalue = False
                for i, val in enumerate(self.closed_array):
                    if val:
                        self.indices.append(i)
            self.boolnotconstant = True

    def set_interpretation_root(self, rep_int_root, recursive = True):
        """Updates the rep_int_root for the polylineset.

        args:
            rep_int_root: new rep_int_root
            recursive: boolean, if true will update individual polys with same root
        """

        self.rep_int_root = rep_int_root

        if recursive:
            for poly in self.polys:
                poly.rep_int_root = rep_int_root

    def convert_to_irap(self, file_name):
        """Output an irap file from a polyline set.

        If file_name exists, it will be overwritten.

        args:
            file_name: output file name for polyline set representation
        """

        end_of_line = np.array([[999.0, 999.0, 999.0]])
        for poly in self.polys:
            if poly == self.polys[0]:
                out_coords = poly.coordinates
            else:
                out_coords = np.concatenate((out_coords, poly.coordinates))
            out_coords = np.concatenate((out_coords, end_of_line))
        np.savetxt(file_name, out_coords, delimiter = ' ')

    def convert_to_charisma(self, file_name):
        """Output to Charisma fault sticks from a polyline set.

        If file_name exists, it will be overwritten.

        args:
            file_name: output file name for polyline set representation
        """

        faultname = self.title.replace(" ", "_") if self.title else 'fault'
        lines = []
        for i, poly in enumerate(self.polys):
            for point in poly.coordinates:
                lines.append(f"INLINE-\t0\t0\t{point[0]}\t{point[1]}\t{point[2]}\t{faultname}\t{i+1}\n")
        with open(file_name, 'w') as f:
            for item in lines:
                f.write(item)

    def extract_property_collection(self, refresh = False):
        """Returns a property collection for the polyline set.

        arguments:
            refresh (bool, default False): if True, the property collection is refreshed
                from the current state of the model; if False and the property collection
                has already been cached for the polyline set, then the cached copy is used

        returns:
            a PropertyCollection holding those properties in the model where this polyline set
            is the supporting representation

        notes:
            polyline set properties may have indexable element of 'nodes' or 'intervals';
            for interval properties, the number of elements per polyline depends on whether the
            polyline is closed or not;
            it is currently left to calling code build the composite property array from data
            for individual polylines, or to find the subset of the combined property data that
            is applicable to an individual polyline
        """
        if self.property_collection is None or refresh:
            self.property_collection = rqp.PropertyCollection(support = self)
        return self.property_collection
